/*******************************************************************************
*
* Name:
*
*   rgdt.c  (Rice GLD Driver Template)
*
* Author:
*
*   Michael Rice               2000
*   RiceM-Drivers@att.net
*
*   This software may be used and distributed according to the terms
*   of the GNU Public License, incorporated herein by reference.
*
* Description:
*   This component provides the network device driver for DEVDESC devices.
*   This driver is a multi-threaded, loadable, clonable, STREAMS hardware 
*   driver supporting the connectionless Data Link Provider Interface, dlpi(7P),
*   over DEVDESC network interface cards.  Multiple DEVDESC controllers
*   installed within the system are supported by the driver.
*
*   This driver provides basic support for the DEVDESC hardware.  Functions
*   including chip initialization, frame transmit and receive, multicast and
*   "promiscuous" support, and error recovery and reporting.
*
*   The driver is dependent on /kernel/misc/gld, a loadable kernel module that
*   provides the driver with the DLPI and STREAMS functionality required of a
*   LAN driver.  See gld(7D) for more details on the primitives supported by
*   the driver.
*
*   The values returned by the driver in the DL_INFO_ACK primitive in response
*   to the DL_INFO_REQ from the user are as follows:
*
*     - The maximum SDU is 1500 (ETHERMTU).
*     - The minimum SDU is 0.  The driver will pad to the mandatory 60-octet
*       minimum packet size.
*     - The DLSAP address length is 8.
*     - The MAC type is DL_ETHER.
*     - The SAP length value is -2, meaning the physical address component is
*       followed immediately by a 2-byte SAP component within the DLSAP address.
*     - The broadcast address value is Ethernet/IEEE broadcast address:
*       (FF:FF:FF:FF:FF:FF).
*
*   Files Relavent to the Driver:
*     /dev/rgdt
*       - Character special device
*     /kernel/drv/rgdt
*       - Driver binary image file
*     /kernel/drv/rgdt.conf
*       - Driver configuration file
*
* Contents:
*
*   Services required (or optional) by all Solaris character mode drivers:
*   _fini                            - Prepare to unload driver.     (External).
*   _info                            - Get driver module info.       (External).
*   _init                            - Initialize driver.            (External).
*   rgdt_aread                       - Read data async from device.  (External).
*                                      Not implemented
*   rgdt_attach                      - Device specific init and PM.  (External).
*   rgdt_awrite                      - Write data async to device.   (External).
*                                      Not implemented
*   rgdt_chpoll                      - Poll device for events.       (External).
*                                      Not implemented
*   rgdt_close                       - Release access to a device.   (External).
*                                      Implemented by GLD.
*   rgdt_detach                      - Device specific term and PM.  (External).
*   rgdt_devmap                      - Translate virtual memory map. (External).
*                                      Not implemented
*   rgdt_getinfo                     - Get device driver info.       (External).
*                                      Implemented by GLD.
*   rgdt_open                        - Get access to a device.       (External).
*                                      Implemented by GLD.
*   rgdt_probe                       - Determine if device present.  (External).
*   rgdt_read                        - Read data from device.        (External).
*                                      Not implemented
*   rgdt_segmap                      - Map device mem to user space. (External).
*                                      Not implemented
*   rgdt_write                       - Write data to device.         (External).
*                                      Not implemented
*
*   Services required (or optional) by Solaris Generic LAN Driver:
*   rgdt_gldm_get_stats              - Get device statistics.        (External).
*   rgdt_gldm_intr                   - Process device interrupt.     (External).
*   rgdt_gldm_ioctl                  - Process I/O control request.  (External).
*                                      Not implemented
*   rgdt_gldm_reset                  - Reset device to initial state.(External).
*   rgdt_gldm_send                   - Send a packet to the device.  (External).
*   rgdt_gldm_set_mac_address        - Set device physical address.  (External).
*   rgdt_gldm_set_multicast          - En/disable mcast reception.   (External).
*   rgdt_gldm_set_promiscuous        - En/disable promiscuouse mode. (External).
*   rgdt_gldm_start                  - Enable the device.            (External).
*   rgdt_gldm_stop                   - Disable the device.           (External).
*
*   General services:
*   rgdt_debug_report                - Report debug information.     (Internal).
*   rgdt_log_message                 - Log a STREAMS message.        (Internal)
*   rgdt_read_properties             - Read driver properties.       (Internal).
*   rgdt_report_error                - Report error information.     (Internal).
*
* Additional Comments:
*
*   1. See the Solaris gld(7D) man page for general information on the 
*      Generic LAN Driver.
*   2. See the Solaris gld(9E) man page for information on the required entry 
*      points for GLD-based device drivers.
*   3. See the Solaris gld(9F) man page for information on the services 
*      provided by GLD for GLD-based device drivers.
*   4. See the Solaris gld_mac_info(9S) and gld_stats(9S) man pages for 
*      information about the GLD-specific data structures used by GLD-based
*      device drivers.
*   5. As of Solaris 8 (SunOS 5.8), the Generic LAN Driver is implemented only
*      for Intel Architecture (IA) systems.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
*******************************************************************************/

/*****************************************************************************/
/*                             Include Files                                 */
/*****************************************************************************/

#include <sys/types.h>                 /* System Type Definitions */
#include <sys/conf.h>                  /* Module Configuration Definitions */
#include <sys/debug.h>                 /* Debug Macro Definitions */
#include <sys/stropts.h>               /* STREAMS Options Definitions */
#include <sys/stream.h>                /* STREAMS Definitions */
#include <sys/strsun.h>                /* Sun STREAMS Definitions */
#include <sys/strlog.h>                /* STREAMS Log Definitions */
#include <sys/dlpi.h>                  /* Data Link Provider Interface Defns */
#include <sys/ethernet.h>              /* Ethernet Definitions */
#include <sys/gld.h>                   /* Generic LAN Driver Definitions */
#include <sys/kmem.h>                  /* Kernel Memory Definitions */
#include <sys/modctl.h>                /* Module Control Definitions */
#include <sys/errno.h>                 /* System Error Code Definitions */
#include <sys/ddi.h>                   /* Device Driver Interface Definitions */
#include <sys/sunddi.h>                /* Sun Device Driver Interface Defns */

#include "rgdt.h"                      /* Driver Definitions */

/*****************************************************************************/
/*                                Pragmas                                    */
/*****************************************************************************/

/*****************************************************************************/
/*                            Module Statics                                 */
/*****************************************************************************/

/* Solaris Driver Entry Point Function Prototypes */

static int rgdt_attach(
  dev_info_t* dev_info_addr,           /* Address of device info */
  ddi_attach_cmd_t command);           /* Command to execute */

static int rgdt_detach(
  dev_info_t* dev_info_addr,           /* Address of device info */
  ddi_detach_cmd_t command);           /* Command to execute */

static int rgdt_probe(
  dev_info_t* dev_info_addr);          /* Address of device info */


/* Generic LAN Driver Entry Point Function Prototypes */

static int rgdt_gldm_get_stats(
  gld_mac_info_t* mac_info_addr,       /* Address of GLD MAC info */
  struct gld_stats* stats_addr);       /* Address of GLD stats info */

static uint_t rgdt_gldm_intr(
  gld_mac_info_t* mac_info_addr);      /* Address of GLD MAC info */

static int rgdt_gldm_reset(
  gld_mac_info_t* mac_info_addr);      /* Address of GLD MAC info */

static int rgdt_gldm_send(
  gld_mac_info_t* mac_info_addr,       /* Address of GLD MAC info */
  mblk_t* message_ptr);                /* Pointer to STREAMS message block */

static int rgdt_gldm_set_mac_address(
  gld_mac_info_t* mac_info_addr,       /* Address of GLD MAC info */
  unsigned char* mac_address_addr);    /* Address of MAC address buffer */

static int rgdt_gldm_set_multicast(
  gld_mac_info_t* mac_info_addr,       /* Address of GLD MAC info */
  unsigned char* mcast_address_addr,   /* Address of MAC address buffer */
  int enable_flag);                    /* Indicate if enabled or disabled */

static int rgdt_gldm_set_promiscuous(
  gld_mac_info_t* mac_info_addr,       /* Address of GLD MAC info */
  int mode_flag);                      /* Which mode, if any, to be enabled */

static int rgdt_gldm_start(
  gld_mac_info_t* mac_info_addr);      /* Address of GLD MAC info */

static int rgdt_gldm_stop(
  gld_mac_info_t* mac_info_addr);      /* Address of GLD MAC info */


/* Module dependency declaration */
char _depends_on[] = "misc/gld";

/* Patchable debugging flag */
/*
 * Note: This value will be overridden if the "debug-level" property is defined
 *       in the driver configuration file (see rgdt_read_properties).
 */
static int rgdt_debug_level = RGDT_DEBUG_LEVEL_0;

/* Ethernet broadcast address */
static rgdt_ether_address_t rgdt_broadcast_address = {
  {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF}};

/* Module information */
static struct module_info rgdt_mod_info = {
  RGDT_MODULE_ID,       /* mi_idnum */
  RGDT_MODULE_NAME,     /* mi_idname */
  RGDT_MIN_PKT_SIZE,    /* mi_minpsz */
  RGDT_MAX_PKT_SIZE,    /* mi_maxpsz */
  RGDT_HI_WATER_MARK,   /* mi_hiwat */
  RGDT_LOW_WATER_MARK}; /* mi_lowat */

/* STREAMS read queue entry points */
static struct qinit rgdt_str_rdinit = {
  (int (*)()) NULL,     /* qi_putp */
  gld_rsrv,             /* qi_srvp */
  gld_open,             /* qi_qopen */
  gld_close,            /* qi_qclose */
  (int (*)()) NULL,     /* qi_qadmin */
  &rgdt_mod_info,       /* qi_minfo */
  NULL};                /* qi_mstat */

/* STREAMS write queue entry points */
static struct qinit rgdt_str_wrinit = {
  gld_wput,             /* qi_putp */
  gld_wsrv,             /* qi_srvp */
  (int (*)()) NULL,     /* qi_qopen */
  (int (*)()) NULL,     /* qi_qclose */
  (int (*)()) NULL,     /* qi_qadmin */
  &rgdt_mod_info,       /* qi_minfo */
  NULL};                /* qi_mstat */

/* STREAMS queue definitions */
static struct streamtab rgdt_str_info = {
  &rgdt_str_rdinit,     /* st_rdinit */
  &rgdt_str_wrinit,     /* st_wrinit */
  NULL,                 /* st_muxrinit */
  NULL};                /* st_muxwrinit */

/* Character/Block entry points definition */
static struct cb_ops rgdt_cb_ops = {
  nulldev,              /* cb_open */
  nulldev,              /* cb_close */
  nodev,                /* cb_strategy */
  nodev,                /* cb_print */
  nodev,                /* cb_dump */
  nodev,                /* cb_read */
  nodev,                /* cb_write */
  nodev,                /* cb_ioctl */
  nodev,                /* cb_devmap */
  nodev,                /* cb_mmap */
  nodev,                /* cb_segmap */
  nochpoll,             /* cb_chpoll */
  ddi_prop_op,          /* cb_prop_op */
  &rgdt_str_info,       /* cb_str */
  D_MP,                 /* cb_flag */
  CB_REV,               /* cb_rev */
  nodev,                /* cb_aread */
  nodev};               /* cb_awrite */

/* Device operations */
static struct dev_ops rgdt_dev_ops = {
  DEVO_REV,             /* devo_rev */
  0,                    /* devo_refcnt */
  gld_getinfo,          /* devo_getinfo */
  nulldev,              /* devo_identify */
  rgdt_probe,           /* devo_probe */
  rgdt_attach,          /* devo_attach */
  rgdt_detach,          /* devo_detach */
  nodev,                /* devo_reset, not implemented in DDI */
  &rgdt_cb_ops,         /* devo_cb_ops */
  NULL,                 /* devo_bus_ops */
  NULL};                /* devo_power */

/* Loadable driver module linkage information */
static struct modldrv rgdt_modldrv = {
  &mod_driverops,          /* Module type (driver) */
  RGDT_MODULE_DESCRIPTION, /* Module description */
  &rgdt_dev_ops};          /* Device operations */

/* Module linkage information */
static struct modlinkage rgdt_modlinkage = {
  MODREV_1, {&rgdt_modldrv, NULL, NULL, NULL}};


/******************************************************************************
*
* Name:
*
*   _fini
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service is a required entry point for all Solaris drivers.  It is 
*   called when the driver is unloaded from the system.
*
*
* Inputs:
*
*   None.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   0         : Module removed successfully.
*   Otherwise : Error removing module.
*
* Additional Comments:
*
*   1. This service is called when the system is shutdown, when the kernel
*      removes the driver from memory, or when the driver is removed using the
*      rem_drv command.
*   2. To prevent the driver from being removed from memory, simply return a 
*      non-zero value.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

int _fini(void)
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;

/*****************************************************************************/
/*                                Begin _fini                                */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT__FINI);

  completion_status = mod_remove(&rgdt_modlinkage);

  RGDT_TRACE_EXIT(RGDT__FINI, completion_status);

  return completion_status;
}  /* End _fini */

/******************************************************************************
*
* Name:
*
*   _info
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service is a required entry point for all Solaris drivers.  It provides
*   information about the loadable module to the kernel.
*
*
* Inputs:
*
*   modinfo_addr - Address of module information.
*
* Outputs:
*
*   modinfo_addr - Address of module information structure filled by mod_info.
*
* Return Values:
*
*   0         : Error getting module information.
*   Otherwise : Module information retrieved successfully.
*
* Additional Comments:
*
*   1. On some Solaris implementations, the _info() service may be called 
*      before the _init() service.  Although this is a bug in Solaris, it does 
*      not affect the driver installation, loading, or initialization.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

int _info(
  struct modinfo* modinfo_addr)        /* Address of module information */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;

/*****************************************************************************/
/*                                Begin _info                                */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT__INFO);

  completion_status = mod_info(&rgdt_modlinkage, modinfo_addr);
  if(completion_status == 0)
  {
    /* Report the error */
    rgdt_report_error(
      NULL,
      RGDT__INFO,
      "Get module info failed",
      RGDT_RPT_WARNING,
      0, completion_status);
  }

  RGDT_TRACE_EXIT(RGDT__INFO, completion_status);

  return completion_status;
}  /* End _info */

/******************************************************************************
*
* Name:
*
*   _init
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service is a required entry point for all Solaris drivers.  It is the 
*   initial entry point into the driver.
*
*
* Inputs:
*
*   None.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   0         : Module installed successfully.
*   Otherwise : Error installing module.
*
* Additional Comments:
*
*   1. On some Solaris implementations, the _info() service may be called 
*      before the _init() service.  Although this is a bug in Solaris, it does 
*      not affect the driver installation, loading, or initialization.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

int _init(void)
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;

/*****************************************************************************/
/*                                Begin _init                                */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT__INIT);

  /* Install the module */
  completion_status = mod_install(&rgdt_modlinkage);
  if(completion_status != 0)
  {
    /* Report the error */
    rgdt_report_error(
      NULL,
      RGDT__INIT,
      "Module installation failed",
      RGDT_RPT_ERROR,
      0, completion_status);
  }

  RGDT_TRACE_EXIT(RGDT__INIT, completion_status);

  return completion_status;
}  /* End _init */

/******************************************************************************
*
* Name:
*
*   rgdt_attach
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service is the device-specific initialization entry point.
*   This service is called once for each instance of the device on the system
*   with command set to DDI_ATTACH.
*
*   This service must create and initialize the GLD MAC info structure and then
*   register the driver with GLD.  All driver initialization must be done prior
*   to registering with GLD; after GLD registration, the driver must be 
*   prepared to handle device interrupts.
*
*
* Inputs:
*
*   dev_info_addr - Address of per-instance device information maintained by OS.
*   command       - Type of attach to perform (attach, resume, etc.)
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   DDI_SUCCESS : Driver initialized and attached to GLD.
*   DDI_FAILURE : The command failed or was not understood.
*
* Additional Comments:
*
*   1. The probe() has already been completed, so we are certain that we are
*      dealing with the correct device type.
*   2. The DDI_PM_RESUME and DDI_RESUME commands are not implemented.
*   3. If the driver does not implement any device-specific IOCTL commands, the
*      gldm_ioctl member of the gld_mac_info_t structure should be NULL.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

static int rgdt_attach(
  dev_info_t* dev_info_addr,           /* Address of device info */
  ddi_attach_cmd_t command)            /* Command to execute */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;
  int svc_status;
  register rgdt_inst_info_addr_t inst_info_addr;
  gld_mac_info_t* gld_info_ptr;

/*****************************************************************************/
/*                             Begin rgdt_attach                             */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_ATTACH);

  if(command == DDI_ATTACH)
  {
    /* Allocate the per-instance information */
    inst_info_addr = (rgdt_inst_info_addr_t)kmem_zalloc(
      sizeof(rgdt_inst_info_t), KM_SLEEP);
    if(inst_info_addr != NULL)
    {
      /* Save the device information */
      inst_info_addr->dev_info_addr = dev_info_addr;

      /* Get the interrupt block cookie */
      svc_status = ddi_get_iblock_cookie(
        dev_info_addr,
        0,
        &(inst_info_addr->iblock_cookie));
      if(svc_status == DDI_SUCCESS)
      {
        /* Setup the non-device-dependent parts of the instance information */
        bcopy(
          &(rgdt_broadcast_address),
          &(inst_info_addr->bcast_address),
          sizeof(rgdt_ether_address_t));
        mutex_init(
          &(inst_info_addr->intr_mutex),
          NULL,
          MUTEX_DRIVER,
          (void*)(inst_info_addr->iblock_cookie));

        /* Set the receive mode flag */
        inst_info_addr->rx_mode = RGDT_RX_STATION | RGDT_RX_BCAST;

        /* Initialize the multicast list */
        inst_info_addr->mcast_list.head_addr = NULL;
        inst_info_addr->mcast_list.count = 0;

        /* Allocate the GLD MAC info for this instance */
        gld_info_ptr = gld_mac_alloc(dev_info_addr);
        if(gld_info_ptr != NULL)
        {
          /* Save the GLD MAC information */
          inst_info_addr->gld_info_ptr = gld_info_ptr;

          /* Do the device-dependent attach */
          /*
           * After this service completes, the hardware is all setup and the 
           * driver is ready to start processing interrupts.
           */
          svc_status = rgdt_attach_dd(inst_info_addr);
          if(svc_status == DDI_SUCCESS)
          {
            /* Add the interrupt handler for the device */
            svc_status = ddi_add_intr(
              dev_info_addr,
              0,
              NULL,
              NULL,
              gld_intr,
              (caddr_t)gld_info_ptr);
            if(svc_status == DDI_SUCCESS)
            {
              /* Setup the GLD MAC info for this instance */
              gld_info_ptr->gldm_private = (caddr_t)inst_info_addr;
              gld_info_ptr->gldm_reset = rgdt_gldm_reset;
              gld_info_ptr->gldm_start = rgdt_gldm_start;
              gld_info_ptr->gldm_stop = rgdt_gldm_stop;
              gld_info_ptr->gldm_set_mac_addr = rgdt_gldm_set_mac_address;
              gld_info_ptr->gldm_set_multicast = rgdt_gldm_set_multicast;
              gld_info_ptr->gldm_set_promiscuous = rgdt_gldm_set_promiscuous;
              gld_info_ptr->gldm_send = rgdt_gldm_send;
              gld_info_ptr->gldm_intr = rgdt_gldm_intr;
              gld_info_ptr->gldm_get_stats = rgdt_gldm_get_stats;
              gld_info_ptr->gldm_ioctl = NULL;
              gld_info_ptr->gldm_ident = "DEVDESC";
              gld_info_ptr->gldm_type = DL_ETHER;
              gld_info_ptr->gldm_minpkt = RGDT_MIN_PKT_SIZE;
              gld_info_ptr->gldm_maxpkt = RGDT_MAX_PKT_SIZE;
              gld_info_ptr->gldm_addrlen = RGDT_ETHER_ADDRESS_LEN;
              gld_info_ptr->gldm_saplen = RGDT_SAP_ADDRESS_LEN;
              gld_info_ptr->gldm_broadcast_addr = inst_info_addr->bcast_address.address;
              gld_info_ptr->gldm_vendor_addr = inst_info_addr->mac_address.address;
              gld_info_ptr->gldm_ppa = ddi_get_instance(dev_info_addr);
              gld_info_ptr->gldm_devinfo = dev_info_addr;
              gld_info_ptr->gldm_cookie = inst_info_addr->iblock_cookie;

              /* Register the driver with GLD */
              svc_status = gld_register(
                dev_info_addr,
                RGDT_MODULE_NAME,
                gld_info_ptr);
              if(svc_status == DDI_SUCCESS)
              {
                completion_status = DDI_SUCCESS;
              }
              else  /* Error registering the driver with GLD */
              {
                /* Report the error */
                rgdt_report_error(
                  inst_info_addr,
                  RGDT_ATTACH,
                  "Register with GLD failed",
                  RGDT_RPT_ERROR,
                  0, svc_status);

                /* Free the per-instance information */
                mutex_destroy(&(inst_info_addr->intr_mutex));
                kmem_free((caddr_t)inst_info_addr, sizeof(rgdt_inst_info_t));

                /* Free the GLD MAC info */
                gld_mac_free(gld_info_ptr);

                completion_status = DDI_FAILURE;
              }
            }
            else  /* Error adding interrupt handler */
            {
              /* Report the error */
              rgdt_report_error(
                inst_info_addr,
                RGDT_ATTACH,
                "Adding interrupt handler failed",
                RGDT_RPT_ERROR,
                0, svc_status);

              /* Free the per-instance information */
              mutex_destroy(&(inst_info_addr->intr_mutex));
              kmem_free((caddr_t)inst_info_addr, sizeof(rgdt_inst_info_t));

              /* Free the GLD MAC info */
              gld_mac_free(gld_info_ptr);

              completion_status = DDI_FAILURE;
            }
          }
          else  /* Error doing device-dependent attach */
          {
            /* Report the error */
            rgdt_report_error(
              inst_info_addr,
              RGDT_ATTACH,
              "Device-dependent attach failed",
              RGDT_RPT_ERROR,
              svc_status, 0);

            /* Free the per-instance information */
            mutex_destroy(&(inst_info_addr->intr_mutex));
            kmem_free((caddr_t)inst_info_addr, sizeof(rgdt_inst_info_t));

            /* Free the GLD MAC info */
            gld_mac_free(gld_info_ptr);

            completion_status = DDI_FAILURE;
          }
        }
        else  /* Error allocating GLD MAC info */
        {
          /* Report the error */
          rgdt_report_error(
            inst_info_addr,
            RGDT_ATTACH,
            "Allocating GLD MAC information failed",
            RGDT_RPT_ERROR,
            0, 0);

          /* Free the per-instance information */
          mutex_destroy(&(inst_info_addr->intr_mutex));
          kmem_free((caddr_t)inst_info_addr, sizeof(rgdt_inst_info_t));

          completion_status = DDI_FAILURE;
        }
      }
      else  /* Error getting interrupt block cookie */
      {
        /* Report the error */
        rgdt_report_error(
          inst_info_addr,
          RGDT_ATTACH,
          "Get interrupt block cookie failed",
          RGDT_RPT_ERROR,
          0, svc_status);

        /* Free the per-instance information */
        kmem_free((caddr_t)inst_info_addr, sizeof(rgdt_inst_info_t));

        completion_status = DDI_FAILURE;
      }
    }
    else  /* Error allocating per-instance information */
    {
      /* Report the error */
      rgdt_report_error(
        NULL,
        RGDT_ATTACH,
        "Allocating per-instance information failed",
        RGDT_RPT_ERROR,
        0, 0);

      completion_status = DDI_FAILURE;
    }
  }
  else if(command == DDI_PM_RESUME)
  {
    /*
     * Note: This command should never be received, since the detach service
     *       does not process DDI_PM_SUSPEND commands.
     */
    /* Report the error */
    rgdt_report_error(
      NULL,
      RGDT_ATTACH,
      "Unimplemented PM resume command",
      RGDT_RPT_WARNING,
      command, 0);

    completion_status = DDI_FAILURE;
  }
  else if(command == DDI_RESUME)
  {
    /*
     * Note: This command should never be received, since the detach service
     *       does not process DDI_SUSPEND commands.
     */
    /* Report the error */
    rgdt_report_error(
      NULL,
      RGDT_ATTACH,
      "Unimplemented resume command",
      RGDT_RPT_WARNING,
      command, 0);

    completion_status = DDI_FAILURE;
  }
  else  /* Unknown command type */
  {
    /* Report the error */
    rgdt_report_error(
      NULL,
      RGDT_ATTACH,
      "Unknown command type",
      RGDT_RPT_WARNING,
      command, 0);

    completion_status = DDI_FAILURE;
  }

  RGDT_TRACE_EXIT(RGDT_ATTACH, completion_status);

  return completion_status;
}  /* End rgdt_attach */

/******************************************************************************
*
* Name:
*
*   rgdt_debug_report
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service may be used by any other function to report information to the
*   system log.  Reporting the message is conditional based on the current
*   value of the "debug level" property.
*
*
* Inputs:
*
*   inst_info_addr   - Address of driver per-instance information.
*   min_debug_level  - Minimum debug level for reporting the message.
*   caller_name_addr - Address of caller/function name.
*   debug_msg_addr   - Address of message to report.
*   msg_type         - Type of message (info, warning, error).
*   app_error_code   - Application error code.
*   os_error_code    - OS error code.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   None.
*
* Additional Comments:
*
*   1. See the comments for the rgdt_report_error() service for information 
*      on where the message data is written to.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

void rgdt_debug_report(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  int min_debug_level,                 /* Minimum debug level */
  char* caller_name_addr,              /* Address of caller name */
  char* debug_msg_addr,                /* Address of debug message */
  uint32_t msg_type,                   /* Message type identifier */
  int app_error_code,                  /* Application error code */
  int os_error_code)                   /* OS error code */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

/*****************************************************************************/
/*                          Begin rgdt_debug_report                          */
/*****************************************************************************/

  /*
   * Check to see if the "debug level" property is at least as high as the 
   * specified minimum.
   */
  if(rgdt_debug_level >= min_debug_level)
  {
    /* Report the message */
    rgdt_report_error(
      inst_info_addr,
      caller_name_addr,
      debug_msg_addr,
      msg_type,
      app_error_code, os_error_code);
  }
}  /* End rgdt_debug_report */

/******************************************************************************
*
* Name:
*
*   rgdt_detach
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service is used to remove the state associated with a given instance
*   of a device node prior to the removal of that instance from the system with
*   the command set to DDI_DETACH.  This service frees all per-instance data,
*   unregisters interrupt handlers and unregisters the driver from the GLD.
*   The device is put into a quiescent state so that it will not generate
*   interrupts.
*
*   This service is called once for each instance of the device for which there
*   has been a successful attach(), once there are no longer any opens on the
*   device.
*
*
* Inputs:
*
*   dev_info_addr - Address of per-instance device information maintained by OS.
*   command       - Type of detach to perform (detach, suspend, etc.)
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   DDI_SUCCESS : The device state was successfully removed or suspended.
*   DDI_FAILURE : The command failed or was not understood.
*
* Additional Comments:
*
*   1. The DDI_PM_SUSPEND and DDI_SUSPEND commands are not implemented.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

static int rgdt_detach(
  dev_info_t* dev_info_addr,           /* Address of device info */
  ddi_detach_cmd_t command)            /* Command to execute */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;
  int svc_status;
  register rgdt_inst_info_addr_t inst_info_addr;
  gld_mac_info_t* gld_info_ptr;
  rgdt_mclist_entry_addr_t mclist_entry_ptr;
  rgdt_mclist_entry_addr_t mclist_prev_entry_ptr;

/*****************************************************************************/
/*                             Begin rgdt_detach                             */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_DETACH);

  /* Get the GLD MAC info */
  gld_info_ptr = (gld_mac_info_t*)ddi_get_driver_private(dev_info_addr);
  if(gld_info_ptr != NULL)
  {
    /* Get the per-instance data from GLD */
    inst_info_addr = (rgdt_inst_info_addr_t)gld_info_ptr->gldm_private;

    /* Check what type of detach command this is */
    if(command == DDI_DETACH)
    {
      /* Unregister the driver from GLD */
      /* Note: The gld_unregister service does the following:
       *       1. Stops the device interrupts, calling gldm_stop if necessary
       *       2. Removes the minor device node
       *       3. Unlinks the driver from GLD
       */
      svc_status = gld_unregister(gld_info_ptr);
      if(svc_status == DDI_SUCCESS)
      {
        /* Do the device-dependent detach */
        svc_status = rgdt_detach_dd(inst_info_addr);
        if(svc_status != DDI_SUCCESS)
        {
          /* Report the error */
          rgdt_report_error(
            inst_info_addr,
            RGDT_DETACH,
            "Device-dependent detach failed",
            RGDT_RPT_ERROR,
            svc_status, 0);
        }

        /* Free the multicast address list */
        if(inst_info_addr->mcast_list.head_addr != NULL)
        {
          mclist_entry_ptr = inst_info_addr->mcast_list.head_addr;
          do
          {
            mclist_prev_entry_ptr = mclist_entry_ptr;
            mclist_entry_ptr = mclist_entry_ptr->next_addr;
            kmem_free(mclist_prev_entry_ptr, sizeof(rgdt_mclist_entry_t));
          } while(mclist_entry_ptr != NULL);
        }

        /* Remove the interrupt handler for the device */
        ddi_remove_intr(
          dev_info_addr,
          0,
          inst_info_addr->iblock_cookie);

        /* Free the mutexes */
        mutex_destroy(&(inst_info_addr->intr_mutex));

        /* Free the per-instance information */
        kmem_free((caddr_t)inst_info_addr, sizeof(rgdt_inst_info_t));

        /* Free the GLD MAC info */
        gld_mac_free(gld_info_ptr);

        completion_status = DDI_SUCCESS;
      }
      else  /* Error unregistering driver from GLD */
      {
        /* Report the error */
        rgdt_report_error(
          inst_info_addr,
          RGDT_DETACH,
          "Unregister from GLD failed",
          RGDT_RPT_ERROR,
          0, svc_status);

        completion_status = DDI_FAILURE;
      }
    }
    else if(command == DDI_PM_SUSPEND)
    {
      /* Report the error */
      rgdt_report_error(
        inst_info_addr,
        RGDT_DETACH,
        "Unimplemented PM suspend command",
        RGDT_RPT_WARNING,
        command, 0);

      completion_status = DDI_FAILURE;
    }
    else if(command == DDI_SUSPEND)
    {
      /* Report the error */
      rgdt_report_error(
        inst_info_addr,
        RGDT_DETACH,
        "Unimplemented suspend command",
        RGDT_RPT_WARNING,
        command, 0);

      completion_status = DDI_FAILURE;
    }
    else  /* Unknown command type */
    {
      /* Report the error */
      rgdt_report_error(
        inst_info_addr,
        RGDT_DETACH,
        "Unknown command type",
        RGDT_RPT_WARNING,
        command, 0);

      completion_status = DDI_FAILURE;
    }
  }
  else  /* Error getting GLD MAC info */
  {
    /* Report the error */
    rgdt_report_error(
      NULL,
      RGDT_DETACH,
      "Getting GLD MAC information failed",
      RGDT_RPT_ERROR,
      0, 0);

    completion_status = DDI_FAILURE;
  }

  RGDT_TRACE_EXIT(RGDT_DETACH, completion_status);

  return completion_status;
}  /* End rgdt_detach */

/******************************************************************************
*
* Name:
*
*   rgdt_gldm_get_stats
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service is a Solaris Generic LAN Driver "callback" function.  It is
*   called by GLD when the driver is required to provide device statistics.
*   It updates the values in the specified GLD statistics structure.
*
*
* Inputs:
*
*   mac_info_addr - Address of driver/device information maintained by GLD.
*   stats_addr    - Address of device statistics information.
*
* Outputs:
*
*   stats_addr - Address of device statistics information.
*
* Return Values:
*
*   GLD_SUCCESS : GLD statistics values were updated.
*
* Additional Comments:
*
*   1. See the Solaris gld(9E) and gld(9F) man pages for more information.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

static int rgdt_gldm_get_stats(
  gld_mac_info_t* mac_info_addr,       /* Address of GLD MAC info */
  struct gld_stats* stats_addr)        /* Address of GLD stats info */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;
  rgdt_inst_info_addr_t inst_info_addr;

/*****************************************************************************/
/*                         Begin rgdt_gldm_get_stats                         */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_GLDM_GET_STATS);

  /* Get the instance information */
  inst_info_addr = (rgdt_inst_info_addr_t)mac_info_addr->gldm_private;

  stats_addr->glds_speed = inst_info_addr->stats.if_speed;
  stats_addr->glds_duplex = inst_info_addr->stats.duplex;
  stats_addr->glds_media = inst_info_addr->stats.media;
  stats_addr->glds_intr = inst_info_addr->stats.intr;
  stats_addr->glds_norcvbuf = inst_info_addr->stats.norcvbuf;
  stats_addr->glds_errrcv = inst_info_addr->stats.ierrors;
  stats_addr->glds_errxmt = inst_info_addr->stats.oerrors;
  stats_addr->glds_missed = inst_info_addr->stats.missed;
  stats_addr->glds_underflow = inst_info_addr->stats.uflo;
  stats_addr->glds_overflow = inst_info_addr->stats.oflo;
  stats_addr->glds_frame = inst_info_addr->stats.align_errors;
  stats_addr->glds_crc = inst_info_addr->stats.fcs_errors;
  stats_addr->glds_nocarrier = inst_info_addr->stats.carrier_errors;
  stats_addr->glds_collisions = inst_info_addr->stats.collisions;
  stats_addr->glds_excoll = inst_info_addr->stats.ex_collisions;
  stats_addr->glds_xmtlatecoll = inst_info_addr->stats.tx_late_collisions;
  stats_addr->glds_defer = inst_info_addr->stats.defer_xmits;
  stats_addr->glds_dot3_first_coll = inst_info_addr->stats.first_collisions;
  stats_addr->glds_dot3_multi_coll = inst_info_addr->stats.multi_collisions;
  stats_addr->glds_dot3_sqe_error = inst_info_addr->stats.sqe_errors;
  stats_addr->glds_dot3_mac_xmt_error = inst_info_addr->stats.macxmt_errors;
  stats_addr->glds_dot3_mac_rcv_error = inst_info_addr->stats.macrcv_errors;
  stats_addr->glds_dot3_frame_too_long = inst_info_addr->stats.toolong_errors;
  stats_addr->glds_short = inst_info_addr->stats.runt_errors;

  completion_status = GLD_SUCCESS;

  RGDT_TRACE_EXIT(RGDT_GLDM_GET_STATS, completion_status);

  return completion_status;
}  /* End rgdt_gldm_get_stats */

/******************************************************************************
*
* Name:
*
*   rgdt_gldm_intr
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service is a Solaris Generic LAN Driver "callback" function.  It is
*   called by GLD when the device _might_ have generated an interrupt.
*   
*
* Inputs:
*
*   mac_info_addr - Address of driver/device information maintained by GLD.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   DDI_INTR_CLAIMED   : The device interrupted and the interrupt was serviced.
*   DDI_INTR_UNCLAIMED : The device did not interrupt.
*
* Additional Comments:
*
*   1. See the Solaris gld(9E) and gld(9F) man pages for more information.
*   2. If the host supports sharing interrupts between devices, the device for 
*      this driver may not be the device that caused the interrupt.
*   3. This service acquires a per-instance mutex.  This prevents simultaneous
*      interrupt processing for multiple instances of the driver.  Acquiring the
*      mutex does not prevent other drivers from processing interrupts.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

static uint_t rgdt_gldm_intr(
  gld_mac_info_t* mac_info_addr)       /* Address of GLD MAC info */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  uint_t completion_status;
  int svc_status;
  rgdt_inst_info_addr_t inst_info_addr;

/*****************************************************************************/
/*                            Begin rgdt_gldm_intr                           */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_GLDM_INTR);

  /* Get the instance information */
  inst_info_addr = (rgdt_inst_info_addr_t)mac_info_addr->gldm_private;

  /*
   * Acquire a mutex to prevent another instance of the driver from entering
   * until the currently executing instance has completed.
   *
   * Note: There is not an issue with the same instance re-entering the mutex.
   *       The device-dependent ISR is required to disable interrupt generation
   *       by the device, if this device is what generated the interrupt.
   *       Device interrupts are re-enabled after this service exits the mutex.
   */
  mutex_enter(&(inst_info_addr->intr_mutex));

  /* Process the interrupt */
  svc_status = rgdt_interrupt_dd(inst_info_addr);
  if(svc_status == DDI_INTR_CLAIMED)
  {
    completion_status = DDI_INTR_CLAIMED;

    /* Exit the mutex */
    mutex_exit(&(inst_info_addr->intr_mutex));

    /* Re-enable interrupts on the device */
    rgdt_enable_interrupts_dd(inst_info_addr);
  }
  else
  {
    completion_status = DDI_INTR_UNCLAIMED;

    /* Exit the mutex */
    mutex_exit(&(inst_info_addr->intr_mutex));
  }

  RGDT_TRACE_EXIT(RGDT_GLDM_INTR, completion_status);

  return completion_status;
}  /* End rgdt_gldm_intr */

/******************************************************************************
*
* Name:
*
*   rgdt_gldm_reset
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service is a Solaris Generic LAN Driver "callback" function.  It is
*   called by GLD to reset the hardware to its initial state.
*
*
* Inputs:
*
*   mac_info_addr - Address of driver/device information maintained by GLD.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   GLD_SUCCESS : The device was reset.
*   GLD_FAILURE : A hardware error occurred.
*
* Additional Comments:
*
*   1. See the Solaris gld(9E) and gld(9F) man pages for more information.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

static int rgdt_gldm_reset(
  gld_mac_info_t* mac_info_addr)       /* Address of GLD MAC info */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;
  int svc_status;
  rgdt_inst_info_addr_t inst_info_addr;

/*****************************************************************************/
/*                           Begin rgdt_gldm_reset                           */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_GLDM_RESET);

  /* Get the instance information */
  inst_info_addr = (rgdt_inst_info_addr_t)mac_info_addr->gldm_private;

  /* Reset the device */
  svc_status = rgdt_reset_dd(inst_info_addr);
  if(svc_status == DDI_SUCCESS)
  {
    completion_status = GLD_SUCCESS;
  }
  else
  {
    /* Report the error */
    rgdt_report_error(
      inst_info_addr,
      RGDT_GLDM_RESET,
      "Device-dependent reset failed",
      RGDT_RPT_ERROR,
      svc_status, 0);

    completion_status = GLD_FAILURE;
  }

  RGDT_TRACE_EXIT(RGDT_GLDM_RESET, completion_status);

  return completion_status;
}  /* End rgdt_gldm_reset */

/******************************************************************************
*
* Name:
*
*   rgdt_gldm_send
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service is a Solaris Generic LAN Driver "callback" function.  It is
*   called by GLD when a packet is available to queue to the device for
*   transmission.
*
*
* Inputs:
*
*   mac_info_addr - Address of driver/device information maintained by GLD.
*   message_ptr   - Pointer to STREAMS message block that contains the data
*                   to be sent to the device.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   GLD_SUCCESS     : The packet was queued/sent successfully.
*   GLD_NORESOURCES : The packet could not be immediately accepted for sending.
*   GLD_BADARG      : An invalid packet or packet length was detected.
*   GLD_FAILURE     : A hardware error occurred.
*
* Additional Comments:
*
*   1. See the Solaris gld(9E) and gld(9F) man pages for more information.
*   2. On success, this service must free the STREAMS message when both the
*      driver and the device no longer require the message (e.g., when the
*      device completes the DMA or the packet has been copied to the device's
*      on-board memory.
*   3. The specified STREAMS message may contain multiple message blocks.
*   4. The specified STREAMS message may contain zero-length message blocks,
*      which should be skipped.
*   5. Prior to sending the packet to the device, the packet length must be 
*      checked to ensure it does not exceed the maximum packet length.
*   6. Prior to sending the packet to the device, it must be padded, 
*      if necessary, to the minimum packet length.
*   7. If the completion status of this service is anything other than success,
*      the STREAMS message must not be free'd.
*   8. If there are no resources available (e.g., the device FIFO is full), 
*      gld_sched() must be called at some later time when there are resources 
*      available.  This will allow GLD to retry the send operation for the 
*      packet.
*   9. If this service returns "no resources" and the gldm_stop() service is
*      called prior to rescheduling the failed packets using gld_sched(), the
*      driver is absolved from the responsibility for rescheduling the packets.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

static int rgdt_gldm_send(
  gld_mac_info_t* mac_info_addr,       /* Address of GLD MAC info */
  mblk_t* message_ptr)                 /* Pointer to STREAMS message block */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;
  int svc_status;
  rgdt_inst_info_addr_t inst_info_addr;

/*****************************************************************************/
/*                            Begin rgdt_gldm_send                           */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_GLDM_SEND);

  /* Get the instance information */
  inst_info_addr = (rgdt_inst_info_addr_t)mac_info_addr->gldm_private;

  /* Check what type of message this is */
  switch(DB_TYPE(message_ptr))
  {
    case M_DATA:
      svc_status = rgdt_transmit_dd(inst_info_addr, message_ptr);
      if(svc_status == DDI_SUCCESS)
      {
        /* Free the message */
        /* Note: This can only be done when neither the driver nor the device
         *       need the message block anymore. This service assumes the 
         *       device-dependent service either made a copy of the message or
         *       has already transferred the data to the device.
         */
        freemsg(message_ptr);

        completion_status = GLD_SUCCESS;
      }
      else  /* No buffer space available to transmit the data */
      {
        /* Do not free the message, it will be tried again later */

        completion_status = GLD_NORESOURCES;
      }
      break;

    default:
      completion_status = GLD_BADARG;
      break;
  }

  RGDT_TRACE_EXIT(RGDT_GLDM_SEND, completion_status);

  return completion_status;
}  /* End rgdt_gldm_send */

/******************************************************************************
*
* Name:
*
*   rgdt_gldm_set_mac_address
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service is a Solaris Generic LAN Driver "callback" function.  It is
*   called by GLD to set the physical address the device should use for 
*   receiving data.  This service programs the device with the specified 
*   address value.
*
*
* Inputs:
*
*   mac_info_addr    - Address of driver/device information maintained by GLD.
*   mac_address_addr - Address of MAC address to write to the device.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   GLD_SUCCESS      : The specified MAC address was written into the device.
*   GLD_NORESOURCES  : Insufficient resources to carry out the request.
*   GLD_NOTSUPPORTED : Programming the device's address is not supported.
*   GLD_BADARG       : An invalid MAC address was specified.
*   GLD_FAILURE      : A hardware error occurred.
*
* Additional Comments:
*
*   1. See the Solaris gld(9E) and gld(9F) man pages for more information.
*   2. This service is usually called by GLD immediately following a 
*      gldm_reset().
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

static int rgdt_gldm_set_mac_address(
  gld_mac_info_t* mac_info_addr,       /* Address of GLD MAC info */
  unsigned char* mac_address_addr)     /* Address of MAC address buffer */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;
  int svc_status;
  rgdt_inst_info_addr_t inst_info_addr;
  rgdt_ether_address_t ether_address;
  char ether_msg[31];

/*****************************************************************************/
/*                      Begin rgdt_gldm_set_mac_address                      */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_GLDM_SET_MAC_ADDRESS);

  /* Get the instance information */
  inst_info_addr = (rgdt_inst_info_addr_t)mac_info_addr->gldm_private;

  /* Create an ethernet address stucture */
  bcopy(mac_address_addr, ether_address.address, RGDT_ETHER_ADDRESS_LEN);

  sprintf(ether_msg, "MAC address: %02x:%02x:%02x:%02x:%02x:%02x",
    ether_address.address[0],
    ether_address.address[1],
    ether_address.address[2],
    ether_address.address[3],
    ether_address.address[4],
    ether_address.address[5]);
  rgdt_debug_report(
    inst_info_addr,
    RGDT_DEBUG_LEVEL_2, RGDT_GLDM_SET_MAC_ADDRESS,
    ether_msg,
    RGDT_RPT_INFO, 0, 0);

  /* Write the MAC address to the device */
  svc_status = rgdt_write_mac_address_dd(
    inst_info_addr,
    &ether_address);
  if(svc_status == DDI_SUCCESS)
  {
    completion_status = DDI_SUCCESS;
  }
  else
  {
    /* Report the error */
    rgdt_report_error(
      inst_info_addr,
      RGDT_GLDM_SET_MAC_ADDRESS,
      "Set MAC address failed",
      RGDT_RPT_ERROR,
      svc_status, 0);

    completion_status = svc_status;
  }

  RGDT_TRACE_EXIT(RGDT_GLDM_SET_MAC_ADDRESS, completion_status);

  return completion_status;
}  /* End rgdt_gldm_set_mac_address */

/******************************************************************************
*
* Name:
*
*   rgdt_gldm_set_multicast
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service is a Solaris Generic LAN Driver "callback" function.  It is
*   called when GLD wants to enable or disable reception of a multicast,
*   group, or functional address.
*
*
* Inputs:
*
*   mac_info_addr      - Address of driver/device information maintained by GLD.
*   mcast_address_addr - Address of multicast address to enable or disable.
*   enable_flag        - Flag that indicates if the specific address should be
*                        enabled or disabled.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   GLD_SUCCESS      : The specified address was enabled/disable.
*   GLD_NORESOURCES  : Insufficient resources to carry out the request.
*   GLD_NOTSUPPORTED : The specified multicast function is not supported.
*   GLD_BADARG       : An invalid address was specified.
*   GLD_FAILURE      : A hardware error occurred.
*
* Additional Comments:
*
*   1. See the Solaris gld(9E) and gld(9F) man pages for more information.
*   2. This service will not be called to enable/disable an address that is
*      already enabled/disabled.
*   3. If disabling an address could result in disabling more than one address
*      at the device level, it is the responsibility of the device driver to
*      keep whatever information is required to avoid disabling an address 
*      that GLD has enabled but not disabled.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

static int rgdt_gldm_set_multicast(
  gld_mac_info_t* mac_info_addr,       /* Address of GLD MAC info */
  unsigned char* mcast_address_addr,   /* Address of MAC address buffer */
  int enable_flag)                     /* Indicate if enabled or disabled */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;
  rgdt_inst_info_addr_t inst_info_addr;
  rgdt_mclist_entry_addr_t mclist_entry_ptr;
  rgdt_mclist_entry_addr_t mclist_prev_entry_ptr;

/*****************************************************************************/
/*                       Begin rgdt_gldm_set_multicast                       */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_GLDM_SET_MULTICAST);

  /* Get the instance information */
  inst_info_addr = (rgdt_inst_info_addr_t)mac_info_addr->gldm_private;

  /* Check if multicast reception should be enabled or disabled */
  if(enable_flag == GLD_MULTI_ENABLE)
  {
    /* Add the address to the multicast list */
    mclist_entry_ptr = (rgdt_mclist_entry_addr_t)kmem_alloc(
      sizeof(rgdt_mclist_entry_t), KM_SLEEP);
    if(mclist_entry_ptr != NULL)
    {
      bcopy(
        mcast_address_addr,
        mclist_entry_ptr->mcast_address.address,
        RGDT_ETHER_ADDRESS_LEN);
      mclist_entry_ptr->next_addr = inst_info_addr->mcast_list.head_addr;
      inst_info_addr->mcast_list.head_addr = mclist_entry_ptr;
      inst_info_addr->mcast_list.count = inst_info_addr->mcast_list.count + 1;

      /* Enable multicast receives */
      inst_info_addr->rx_mode = inst_info_addr->rx_mode | RGDT_RX_MCAST;

      /* Update the receive mode */
      rgdt_set_rx_mode_dd(inst_info_addr);

      completion_status = GLD_SUCCESS;
    }
    else  /* Error allocating memory for the new multicast address list entry */
    {
      rgdt_report_error(
        inst_info_addr,
        RGDT_GLDM_SET_MULTICAST,
        "Memory allocation failure for list entry",
        RGDT_RPT_ERROR,
        0, 0);

      completion_status = GLD_NORESOURCES;
    }
  }
  else if(enable_flag == GLD_MULTI_DISABLE)
  {
    /* Find the specified address on the multicast list */
    mclist_entry_ptr = inst_info_addr->mcast_list.head_addr;
    mclist_prev_entry_ptr = NULL;
    while(mclist_entry_ptr)
    {
      if(bcmp(
        mclist_entry_ptr->mcast_address.address,
        mcast_address_addr,
        RGDT_ETHER_ADDRESS_LEN) == 0)
      {
        break;
      }
      else
      {
        mclist_prev_entry_ptr = mclist_entry_ptr;
        mclist_entry_ptr = mclist_entry_ptr->next_addr;
      }
    }

    /* Check if the specified address was found on the multicast list */
    if(mclist_entry_ptr)
    {
      /* Remove the address from the multicast list */
      if(inst_info_addr->mcast_list.head_addr == mclist_entry_ptr)
      {
        inst_info_addr->mcast_list.head_addr = mclist_entry_ptr->next_addr;
      }
      else
      {
        mclist_prev_entry_ptr->next_addr = mclist_entry_ptr->next_addr;
      }
      kmem_free(mclist_entry_ptr, sizeof(rgdt_mclist_entry_t));

      inst_info_addr->mcast_list.count = inst_info_addr->mcast_list.count - 1;

      /* Check if multicast receives should be disabled */
      if(inst_info_addr->mcast_list.count == 0)
      {
        inst_info_addr->rx_mode = inst_info_addr->rx_mode & ~RGDT_RX_MCAST;
      }

      /* Update the receive mode */
      rgdt_set_rx_mode_dd(inst_info_addr);

      completion_status = GLD_SUCCESS;
    }
    else  /* Address not found on multicast list */
    {
      /* Report the error */
      rgdt_report_error(
        inst_info_addr,
        RGDT_GLDM_SET_MULTICAST,
        "Specified address not disabled, not found on list",
        RGDT_RPT_ERROR,
        0, 0);

      completion_status = GLD_BADARG;
    }
  }
  else
  {
    completion_status = GLD_NOTSUPPORTED;
  }

  RGDT_TRACE_EXIT(RGDT_GLDM_SET_MULTICAST, completion_status);

  return completion_status;
}  /* End rgdt_gldm_set_multicast */

/******************************************************************************
*
* Name:
*
*   rgdt_gldm_set_promiscuous
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service is a Solaris Generic LAN Driver "callback" function.  It is
*   called by GLD to enable or disable "promiscuous" mode on the device.  This
*   enables the device to receive all packets or multicast packets on the 
*   medium.
*
*
* Inputs:
*
*   mac_info_addr - Address of driver/device information maintained by GLD.
*   mode_flag     - Flag that specifies the promiscuous mode to be used.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   GLD_SUCCESS      : The specified promiscuous mode was set successfully.
*   GLD_NORESOURCES  : Insufficient resources to carry out the request.
*   GLD_NOTSUPPORTED : The specified promiscuous mode is not supported.
*   GLD_FAILURE      : A hardware error occurred.
*
* Additional Comments:
*
*   1. See the Solaris gld(9E) and gld(9F) man pages for more information.
*   2. If the specified mode is to enable promiscuous multicast, but the 
*      device does not support multicast-only promiscuous mode, the driver
*      must enable physical promiscuous mode (i.e., receive all packets);
*      GLD will filter out any superfluous packets.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

static int rgdt_gldm_set_promiscuous(
  gld_mac_info_t* mac_info_addr,       /* Address of GLD MAC info */
  int mode_flag)                       /* Which mode, if any, to be enabled */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;
  rgdt_inst_info_addr_t inst_info_addr;

/*****************************************************************************/
/*                      Begin rgdt_gldm_set_promiscuous                      */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_GLDM_SET_PROMISCUOUS);

  /* Get the instance information */
  inst_info_addr = (rgdt_inst_info_addr_t)mac_info_addr->gldm_private;

  /* Check what promiscuous mode to set */
  switch(mode_flag)
  {
    case GLD_MAC_PROMISC_MULTI:
      inst_info_addr->rx_mode = 
        inst_info_addr->rx_mode | RGDT_RX_PRM_MCAST;
      break;

    case GLD_MAC_PROMISC_NONE:
      inst_info_addr->rx_mode = 
        inst_info_addr->rx_mode & ~(RGDT_RX_PRM | RGDT_RX_PRM_MCAST);
      break;

    case GLD_MAC_PROMISC_PHYS:
    default:
      inst_info_addr->rx_mode = 
        inst_info_addr->rx_mode | RGDT_RX_PRM;
      break;
  }

  /* Update the receive mode */
  rgdt_set_rx_mode_dd(inst_info_addr);

  completion_status = GLD_SUCCESS;

  RGDT_TRACE_EXIT(RGDT_GLDM_SET_PROMISCUOUS, completion_status);

  return completion_status;
}  /* End rgdt_gldm_set_promiscuous */

/******************************************************************************
*
* Name:
*
*   rgdt_gldm_start
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service is a Solaris Generic LAN Driver "callback" function.  It is
*   called by GLD to enable the device to generate interrupts and prepare the
*   driver to receive data from the device.
*
*
* Inputs:
*
*   mac_info_addr - Address of driver/device information maintained by GLD.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   GLD_SUCCESS : The operation completed successfully.
*   GLD_FAILURE : A hardware error occurred.
*
* Additional Comments:
*
*   1. See the Solaris gld(9E) and gld(9F) man pages for more information.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

static int rgdt_gldm_start(
  gld_mac_info_t* mac_info_addr)       /* Address of GLD MAC info */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;
  int svc_status;
  rgdt_inst_info_addr_t inst_info_addr;

/*****************************************************************************/
/*                           Begin rgdt_gldm_start                           */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_GLDM_START);

  /* Get the instance information */
  inst_info_addr = (rgdt_inst_info_addr_t)mac_info_addr->gldm_private;

  /* Start the device */
  svc_status = rgdt_start_dd(inst_info_addr);
  if(svc_status == DDI_SUCCESS)
  {
    completion_status = GLD_SUCCESS;
  }
  else
  {
    /* Report the error */
    rgdt_report_error(
      inst_info_addr,
      RGDT_GLDM_START,
      "Device-dependent start failed",
      RGDT_RPT_ERROR,
      svc_status, 0);

    completion_status = GLD_FAILURE;
  }

  RGDT_TRACE_EXIT(RGDT_GLDM_START, completion_status);

  return completion_status;
}  /* End rgdt_gldm_start */

/******************************************************************************
*
* Name:
*
*   rgdt_gldm_stop
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service is a Solaris Generic LAN Driver "callback" function.  It is
*   called by GLD to disable the device from generating interrupts and stop
*   the driver from calling gld_recv() for received data.
*
*
* Inputs:
*
*   mac_info_addr - Address of driver/device information maintained by GLD.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   GLD_SUCCESS : Always returned, regardless of the completion status.
*
* Additional Comments:
*
*   1. See the Solaris gld(9E) and gld(9F) man pages for more information.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

static int rgdt_gldm_stop(
  gld_mac_info_t* mac_info_addr)       /* Address of GLD MAC info */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;
  int svc_status;
  rgdt_inst_info_addr_t inst_info_addr;

/*****************************************************************************/
/*                           Begin rgdt_gldm_stop                            */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_GLDM_STOP);

  /* Get the instance information */
  inst_info_addr = (rgdt_inst_info_addr_t)mac_info_addr->gldm_private;

  /* Stop the device */
  svc_status = rgdt_stop_dd(inst_info_addr);
  if(svc_status == DDI_SUCCESS)
  {
  }
  else
  {
    /* Report the error */
    rgdt_report_error(
      inst_info_addr,
      RGDT_GLDM_STOP,
      "Device-dependent stop failed",
      RGDT_RPT_ERROR,
      svc_status, 0);
  }

  completion_status = GLD_SUCCESS;

  RGDT_TRACE_EXIT(RGDT_GLDM_STOP, completion_status);

  return completion_status;
}  /* End rgdt_gldm_stop */

/******************************************************************************
*
* Name:
*
*   rgdt_log_message
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service writes the data in the specified STREAMS message to the system
*   log file, using the debug reporting services.
*
*
* Inputs:
*
*   description_addr - Address of buffer with description of the message being 
*                      logged, usually the calling service's name.
*   message_addr     - Address of STREAMS message with data to log.
*
* Outputs:
*
*   Data from STREAMS message written to system log file.
*
* Return Values:
*
*   None.
*
* Additional Comments:
*
*   1. This service is for debugging purposes.  The driver parameter
*      "debug-level", specified in the driver configuration file, must be
*      set to 3 or higher for the data to be written to the log.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

void rgdt_log_message(
  char* description_addr,              /* Address of description of message */
  mblk_t* message_addr)                /* Address of STREAMS message to log */
{
/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  register int message_len;
  mblk_t* block_ptr;
  char debug_msg[1024];

/*****************************************************************************/
/*                           Begin rgdt_log_message                          */
/*****************************************************************************/

  /*
   * Check the current debug level so we don't waste time if the data is not
   * going to be written to the log.
   */
  if(rgdt_debug_level >= RGDT_DEBUG_LEVEL_3)
  {
    /* Get the message length */
    message_len = 0;
    for(block_ptr = message_addr; 
        block_ptr != NULL; 
        block_ptr = block_ptr->b_cont)
    {
      message_len += MBLKL(block_ptr);
    }

    sprintf(debug_msg, "%s (message length = 0x%02X bytes)",
      description_addr,
      message_len);
    rgdt_report_error(
      NULL, RGDT_LOG_MESSAGE, debug_msg, RGDT_RPT_INFO, 0, 0);

    /* TBD: Add code to dump the message data to the log. */
  }  /* End if debug level is high enough */
}  /* End rgdt_log_message */

/******************************************************************************
*
* Name:
*
*   rgdt_probe
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service determines if the device specified in the device information
*   actually exists and is a valid device for this driver.  This service is
*   called after identify() and before attach() for a given device.
*
*   This service only probes the device.  It does not create or change any
*   software state.  Device initialization is performed in attach().
*
*   For a self-identifying device, this service is not necessary.  However, if
*   a device exists in both self-identifying and non-self-identifying forms,
*   this service can be used to simplify the driver.  The ddi_dev_is_sid()
*   service can be used to determine if this service needs to do any work.
*
*
* Inputs:
*
*   dev_info_addr - Address of per-instance device information maintained by OS.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   DDI_PROBE_SUCCESS  : Probe was successful.
*   DDI_PROBE_FAILURE  : The probe failed.
*   DDI_PROBE_DONTCARE : Probe was unsuccessful, but attach() should be called.
*   DDI_PROBE_PARTIAL  : Device not present now, but may be in the future.
*
* Additional Comments:
*
*   1. The identify() service is obsolete and is not defined for new drivers.
*   2. The rgdt_read_properties() service is called here becuase this is the
*      first entry point that has the Solaris Device Information available.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

static int rgdt_probe(
  dev_info_t* dev_info_addr)           /* Address of device info */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;

/*****************************************************************************/
/*                              Begin rgdt_probe                             */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_PROBE);

  /* Read the driver properties */
  rgdt_read_properties(dev_info_addr);

  /* Probe to see if the device is actually there */
  switch(rgdt_probe_dd(dev_info_addr))
  {
    case DDI_PROBE_SUCCESS:
      /* The device is actually in the system */
      completion_status = DDI_PROBE_SUCCESS;
      break;

    case DDI_PROBE_DONTCARE:
      /* The device may be in the system */
      completion_status = DDI_PROBE_DONTCARE;
      break;

    case DDI_PROBE_PARTIAL:
      /* The device is not in the system, but may be in the future */
      completion_status = DDI_PROBE_PARTIAL;
      break;

    case DDI_FAILURE:
    default:
      /* The device is not in the system */
      completion_status = DDI_PROBE_FAILURE;
      break;
  }

  RGDT_TRACE_EXIT(RGDT_PROBE, completion_status);

  return completion_status;
}  /* End rgdt_probe */

/******************************************************************************
*
* Name:
*
*   rgdt_read_properties
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service reads all of the properties defined in the device
*   configuration file.  The properties read are global for the driver.
*
*
* Inputs:
*
*   dev_info_addr - Address of per-instance device information maintained by OS.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   None.
*
* Additional Comments:
*
*   1. The driver configuration file is colocated with the driver image file.
*   2. It is possible to set a static global driver variable from the 
*      /etc/system file (e.g., set mydrv:drvvar=xxx).
*      If that global driver variable also correlates to a driver property that
*      is read by this service, then the property value will override the 
*      value defined in /etc/system.  However, since this service is not called
*      until well into the driver loading/configuration stage, setting the 
*      variable in /etc/system may still have a useful result.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

void rgdt_read_properties(
  dev_info_t* dev_info_addr)           /* Address of device info */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int int_prop_value;

/*****************************************************************************/
/*                         Begin rgdt_read_properties                        */
/*****************************************************************************/

  /*
   * Read the debug-level property, don't override the current setting if the
   * property is not defined.
   */
  int_prop_value = ddi_prop_get_int(
    DDI_DEV_T_ANY,
    dev_info_addr,
    DDI_PROP_DONTPASS,
    "debug-level", -1);
  if(int_prop_value != -1)
  {
    rgdt_debug_level = int_prop_value;
  }
}  /* End rgdt_read_properties */

/******************************************************************************
*
* Name:
*
*   rgdt_report_error
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service may be used by any other function to report information to the
*   system log.  This service also provides the capability to cause a system
*   panic, if the caller requests it.
*
*
* Inputs:
*
*   inst_info_addr   - Address of driver per-instance information.
*   caller_name_addr - Address of caller/function name.
*   error_msg_addr   - Address of message to report.
*   msg_type         - Type of message (info, warning, error).
*   app_error_code   - Application error code.
*   os_error_code    - OS error code.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   None.
*
* Additional Comments:
*
*   1. Messages are written to the system log file.  If the kernel is booted
*      with the "verbose" flag (i.e., boot -v), then the messages are also
*      displayed on the system console.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

void rgdt_report_error(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  char* caller_name_addr,              /* Address of caller name */
  char* error_msg_addr,                /* Address of error message */
  uint32_t msg_type,                   /* Message type identifier */
  int app_error_code,                  /* Application error code */
  int os_error_code)                   /* OS error code */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int rpt_level;
  char* rpt_type_names[] = {
    "",
    "WARNING: ",
    "ERROR: "};
  int type_name_index;

/*****************************************************************************/
/*                          Begin rgdt_report_error                          */
/*****************************************************************************/

  /* Check to see if the system should panic */
  if(msg_type & RGDT_RPT_FATAL)
  {
    rpt_level = CE_PANIC;
    type_name_index = 2;
  }
  else  /* No panic necessary */
  {
    rpt_level = CE_CONT;
    switch(msg_type & RGDT_RPT_MSG_TYPE_MASK)
    {
      case RGDT_RPT_INFO:
        type_name_index = 0;
        break;
      case RGDT_RPT_WARNING:
        type_name_index = 1;
        break;
      case RGDT_RPT_ERROR:
      default:
        type_name_index = 2;
        break;
    }
  }

  /* Report the message */
  if(caller_name_addr != NULL)
  {
    if(inst_info_addr != NULL)
    {
      cmn_err(
        rpt_level,
        "?%s%s%d(%s) - %s, App code: 0x%x, OS code: 0x%x\n",
        rpt_type_names[type_name_index],
        ddi_get_name(inst_info_addr->dev_info_addr),
        ddi_get_instance(inst_info_addr->dev_info_addr),
        caller_name_addr,
        error_msg_addr,
        app_error_code,
        os_error_code);
    }
    else
    {
      cmn_err(
        rpt_level,
        "?%s%s?(%s) - %s, App code: 0x%x, OS code: 0x%x\n",
        rpt_type_names[type_name_index],
        RGDT_MODULE_NAME,
        caller_name_addr,
        error_msg_addr,
        app_error_code,
        os_error_code);
    }
  }
  else
  {
    if(inst_info_addr != NULL)
    {
      cmn_err(
        rpt_level,
        "?%s%s%d - %s, App code: 0x%x, OS code: 0x%x\n",
        rpt_type_names[type_name_index],
        ddi_get_name(inst_info_addr->dev_info_addr),
        ddi_get_instance(inst_info_addr->dev_info_addr),
        error_msg_addr,
        app_error_code,
        os_error_code);
    }
    else
    {
      cmn_err(
        rpt_level,
        "?%s%s? - %s, App code: 0x%x, OS code: 0x%x\n",
        rpt_type_names[type_name_index],
        RGDT_MODULE_NAME,
        error_msg_addr,
        app_error_code,
        os_error_code);
    }
  }
}  /* End rgdt_report_error */
