/*******************************************************************************
*
* Name:
*
*   rgdt.h  (Rice GLD Driver Template)
*
* Author:
*
*   Michael Rice               2000
*   RiceM-Drivers@att.net
*
*   This software may be used and distributed according to the terms
*   of the GNU Public License, incorporated herein by reference.
*
* Description:
*   This component provides the network device driver for DEVDESC
*   devices.  This driver is a multi-threaded, loadable, clonable, STREAMS 
*   hardware driver supporting the connectionless Data Link Provider Interface,
*   dlpi(7P), over DEVDESC network interface cards.  Multiple DEVDESC 
*   controllers installed within the system are supported by the driver.
*
*   This driver provides basic support for the DEVDESC hardware.  Functions
*   including chip initialization, frame transmit and receive, multicast and
*   "promiscuous" support, and error recovery and reporting.
*
*   The driver is dependent on /kernel/misc/gld, a loadable kernel module that
*   provides the driver with the DLPI and STREAMS functionality required of a
*   LAN driver.  See gld(7D) for more details on the primitives supported by
*   the driver.
*
*   The values returned by the driver in the DL_INFO_ACK primitive in response
*   to the DL_INFO_REQ from the user are as follows:
*
*     - The maximum SDU is 1500 (ETHERMTU).
*     - The minimum SDU is 0.  The driver will pad to the mandatory 60-octet
*       minimum packet size.
*     - The DLSAP address length is 8.
*     - The MAC type is DL_ETHER.
*     - The SAP length value is -2, meaning the physical address component is
*       followed immediately by a 2-byte SAP component within the DLSAP address.
*     - The broadcast address value is Ethernet/IEEE broadcast address:
*       (FF:FF:FF:FF:FF:FF).
*
*   Files Relavent to the Driver:
*     /dev/rgdt
*       - Character special device
*     /kernel/drv/rgdt
*       - Driver binary image file
*     /kernel/drv/rgdt.conf
*       - Driver configuration file
*
* Contents:
*
*   Services required (or optional) by all Solaris character mode drivers:
*   _fini                            - Prepare to unload driver.     (External).
*   _info                            - Get driver module info.       (External).
*   _init                            - Initialize driver.            (External).
*   rgdt_aread                       - Read data async from device.  (External).
*                                      Not implemented
*   rgdt_attach                      - Device specific init and PM.  (External).
*   rgdt_awrite                      - Write data async to device.   (External).
*                                      Not implemented
*   rgdt_chpoll                      - Poll device for events.       (External).
*                                      Not implemented
*   rgdt_close                       - Release access to a device.   (External).
*                                      Implemented by GLD.
*   rgdt_detach                      - Device specific term and PM.  (External).
*   rgdt_devmap                      - Translate virtual memory map. (External).
*                                      Not implemented
*   rgdt_getinfo                     - Get device driver info.       (External).
*                                      Implemented by GLD.
*   rgdt_open                        - Get access to a device.       (External).
*                                      Implemented by GLD.
*   rgdt_probe                       - Determine if device present.  (External).
*   rgdt_read                        - Read data from device.        (External).
*                                      Not implemented
*   rgdt_segmap                      - Map device mem to user space. (External).
*                                      Not implemented
*   rgdt_write                       - Write data to device.         (External).
*                                      Not implemented
*
*   Services required (or optional) by Solaris Generic LAN Driver:
*   rgdt_gldm_get_stats              - Get device statistics.        (External).
*   rgdt_gldm_intr                   - Process device interrupt.     (External).
*   rgdt_gldm_ioctl                  - Process I/O control request.  (External).
*                                      Not implemented
*   rgdt_gldm_reset                  - Reset device to initial state.(External).
*   rgdt_gldm_send                   - Send a packet to the device.  (External).
*   rgdt_gldm_set_mac_address        - Set device physical address.  (External).
*   rgdt_gldm_set_multicast          - En/disable mcast reception.   (External).
*   rgdt_gldm_set_promiscuous        - En/disable promiscuouse mode. (External).
*   rgdt_gldm_start                  - Enable the device.            (External).
*   rgdt_gldm_stop                   - Disable the device.           (External).
*
*   Device-dependent services:
*   rgdt_attach_dd                   - Device-dependent attach.      (Internal).
*   rgdt_detach_dd                   - Device-dependent detach.      (Internal).
*   rgdt_disable_interrupts_dd       - Disable interrupt delivery.   (Internal).
*   rgdt_do_mii_autoneg_dd           - Auto-negotiate a link.        (Internal).
*   rgdt_enable_interrupts_dd        - Enable interrupt delivery.    (Internal).
*   rgdt_interrupt_dd                - Process an interrupt.         (Internal).
*   rgdt_mdio_read_dd                - Read MII values.              (Internal).
*   rgdt_mdio_sync_dd                - Generate MII sync preamble.   (Internal).
*   rgdt_mdio_write_dd               - Write MII values.             (Internal).
*   rgdt_probe_dd                    - Device-dependent probe.       (Internal).
*   rgdt_read_mac_address_dd         - Read MAC address from device. (Internal).
*   rgdt_reset_dd                    - Device-dependent reset.       (Internal).
*   rgdt_rx_interrupt_dd             - Process a receive interrupt.  (Internal).
*   rgdt_set_rx_mode_dd              - Set the device receive mode.  (Internal).
*   rgdt_start_dd                    - Device-dependent start.       (Internal).
*   rgdt_stop_dd                     - Device-dependent stop.        (Internal).
*   rgdt_transmit_dd                 - Transmit data to the device.  (Internal).
*   rgdt_tx_interrupt_dd             - Process a transmit interrupt. (Internal).
*   rgdt_write_mac_address_dd        - Write MAC address to device.  (Internal).
*
*   General services:
*   rgdt_debug_report                - Report debug information.     (Internal).
*   rgdt_log_message                 - Log a STREAMS message.        (Internal)
*   rgdt_read_properties             - Read driver properties.       (Internal).
*   rgdt_report_error                - Report error information.     (Internal).
*
* Additional Comments:
*
*   1. See the Solaris gld(7D) man page for general information on the 
*      Generic LAN Driver.
*   2. See the Solaris gld(9E) man page for information on the required entry 
*      points for GLD-based device drivers.
*   3. See the Solaris gld(9F) man page for information on the services 
*      provided by GLD for GLD-based device drivers.
*   4. See the Solaris gld_mac_info(9S) and gld_stats(9S) man pages for 
*      information about the GLD-specific data structures used by GLD-based
*      device drivers.
*   5. As of Solaris 8 (SunOS 5.8), the Generic LAN Driver is implemented only
*      for Intel Architecture (IA) systems.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
*******************************************************************************/

#if !defined(_RGDT_H_)
#define _RGDT_H_

/*****************************************************************************/
/*                             Include Files                                 */
/*****************************************************************************/

/* Standard C Header File */

#if defined(__cplusplus)
extern "C" {
#endif

/******************************************************************************/
/*                              Constant Definitions                          */
/******************************************************************************/

/* General Definitions */

#define RGDT_SUCCESS                1
#define RGDT_FAILURE                0

#define RGDT_MODULE_NAME            "rgdt"    /* Module name */
#define RGDT_MODULE_DESCRIPTION     "MODDESC"
#define	RGDT_MODULE_ID              (68)      /* Module ID number */
#define RGDT_ETHER_ADDRESS_LEN      (6)       /* Ethernet address length */
#define RGDT_SAP_ADDRESS_LEN        (-2)      /* Service Access Point len */
#define RGDT_MIN_ETHER_SIZE         (60)      /* Min frame size w/o CRC */
#define RGDT_MAX_ETHER_SIZE         (1514)    /* Max frame size w/o CRC */
#define RGDT_MIN_PKT_SIZE           (0)       /* Min packet size */
#define RGDT_MAX_PKT_SIZE           (1500)    /* Max packet size */
#define RGDT_LOW_WATER_MARK         (1)       /* Low-water mark */
#define RGDT_HI_WATER_MARK          (32*1024) /* Hi-water mark */

#define RGDT_MAX_ISR_EVENTS         128       /* Max events in ISR */

/* Debug Level Definitions */

#define RGDT_DEBUG_LEVEL_0          (0)       /* No debug information */
#define RGDT_DEBUG_LEVEL_1          (1)       /* Minimum debug information */
#define RGDT_DEBUG_LEVEL_2          (2)       /* Medium debug information */
#define RGDT_DEBUG_LEVEL_3          (3)       /* Maximum debug information */

/* Error Reporting Message Types */

#define RGDT_RPT_INFO               0x1      /* Informational message */
#define RGDT_RPT_WARNING            0x2      /* Warning message */
#define RGDT_RPT_ERROR              0x4      /* Error message */
#define RGDT_RPT_MSG_TYPE_MASK      0x7      /* Mask to extract type */
#define RGDT_RPT_FATAL              0x8      /* Message is fatal, panic */

/* Receive Mode Definitions */

#define RGDT_RX_STATION             0x01      /* Receive station packets */
#define RGDT_RX_MCAST               0x02      /* Receive mcast packets */
#define RGDT_RX_BCAST               0x04      /* Receive bcast packets */
#define RGDT_RX_PRM                 0x08      /* Receive all packets */
#define RGDT_RX_PRM_MCAST           0x10      /* Receive all mcast packets */

/* Media Selection Definitions */

#define RGDT_DUPLEX_AUTO            0         /* Auto-negotiate duplex */
#define RGDT_DUPLEX_HALF            1         /* Force half-duplex */
#define RGDT_DUPLEX_FULL            2         /* Force full-duplex */
#define RGDT_SPEED_AUTO             0         /* Auto-negotiate speed */
#define RGDT_SPEED_10               10        /* Force 10 Mbps speed */
#define RGDT_SPEED_100              100       /* Force 100 Mbps speed */

#define RGDT_AUTO_NEGOTIATE_TIMEOUT 5000000   /* Timeout for autoneg (us) */

/* PCI Configuration Space Definitions */

#define RGDT_PCICFG_VENDOR_ID        0x00
#define RGDT_PCICFG_DEVICE_ID        0x02
#define RGDT_PCICFG_CMD              0x04
#define RGDT_PCICFG_STATUS           0x06
#define RGDT_PCICFG_REVISION_ID      0x08
#define RGDT_PCICFG_SUBCLASS_CODE    0x09
#define RGDT_PCICFG_CLASS_CODE       0x0B
#define RGDT_PCICFG_CACHE_LINE_SIZE  0x0C
#define RGDT_PCICFG_LATENCY_TIMER    0x0D
#define RGDT_PCICFG_HEADER_TYPE      0x0E
#define RGDT_PCICFG_IO_BASE_ADDRESS  0x10
#define RGDT_PCICFG_MEM_BASE_ADDRESS 0x14
#define RGDT_PCICFG_SUBSYS_VENDOR_ID 0x2C
#define RGDT_PCICFG_SUBSYS_ID        0x2E
#define RGDT_PCICFG_BIOS             0x30
#define RGDT_PCICFG_INT_LINE         0x3C
#define RGDT_PCICFG_INT_PIN          0x3D
#define RGDT_PCICFG_MIN_GRANT        0x3E
#define RGDT_PCICFG_MAX_GRANT        0x3F

/* PCI Vendor and Device ID */

#define RGDT_VENDOR_ID              DEVVEND   /* PCI vendor ID */
#define RGDT_DEVICE_ID              DEVID     /* PCI device ID */

/* Register Offsets and Operations */

/* TBD: Define the register offsets, device commands, etc. */


/* Service Name Definitions */

#define RGDT__FINI                  "_fini"
#define RGDT__INFO                  "_info"
#define RGDT__INIT                  "_init"
#define RGDT_ATTACH                 "rgdt_attach"
#define RGDT_ATTACH_DD              "rgdt_attach_dd"
#define RGDT_DEBUG_REPORT           "rgdt_debug_report"
#define RGDT_DETACH                 "rgdt_detach"
#define RGDT_DETACH_DD              "rgdt_detach_dd"
#define RGDT_DISABLE_INTERRUPTS_DD  "rgdt_disable_interrupts_dd"
#define RGDT_DO_MII_AUTONEG_DD      "rgdt_do_mii_autoneg_dd"
#define RGDT_ENABLE_INTERRUPTS_DD   "rgdt_enable_interrupts_dd"
#define RGDT_GLDM_GET_STATS         "rgdt_gldm_get_stats"
#define RGDT_GLDM_INTR              "rgdt_gldm_intr"
#define RGDT_GLDM_RESET             "rgdt_gldm_reset"
#define RGDT_GLDM_SEND              "rgdt_gldm_send"
#define RGDT_GLDM_SET_MAC_ADDRESS   "rgdt_gldm_set_mac_address"
#define RGDT_GLDM_SET_MULTICAST     "rgdt_gldm_set_multicast"
#define RGDT_GLDM_SET_PROMISCUOUS   "rgdt_gldm_set_promiscuous"
#define RGDT_GLDM_START             "rgdt_gldm_start"
#define RGDT_GLDM_STOP              "rgdt_gldm_stop"
#define RGDT_INTERRUPT_DD           "rgdt_interrupt_dd"
#define RGDT_LOG_MESSAGE            "rgdt_log_message"
#define RGDT_MDIO_READ_DD           "rgdt_mdio_read_dd"
#define RGDT_MDIO_WRITE_DD          "rgdt_mdio_write_dd"
#define RGDT_PROBE                  "rgdt_probe"
#define RGDT_PROBE_DD               "rgdt_probe_dd"
#define RGDT_READ_EEPROM_DD         "rgdt_read_eeprom_dd"
#define RGDT_READ_MAC_ADDRESS_DD    "rgdt_read_mac_address_dd"
#define RGDT_READ_PROPERTIES        "rgdt_read_properties"
#define RGDT_REPORT_ERROR           "rgdt_report_error"
#define RGDT_RESET_DD               "rgdt_reset_dd"
#define RGDT_RX_INTERRUPT_DD        "rgdt_rx_interrupt_dd"
#define RGDT_SET_RX_MODE_DD         "rgdt_set_rx_mode_dd"
#define RGDT_START_DD               "rgdt_start_dd"
#define RGDT_STOP_DD                "rgdt_stop_dd"
#define RGDT_TRANSMIT_DD            "rgdt_transmit_dd"
#define RGDT_TX_INTERRUPT_DD        "rgdt_tx_interrupt_dd"
#define RGDT_WRITE_MAC_ADDRESS_DD   "rgdt_write_mac_address_dd"

/* Macro Definitions */

#define RGDT_TRACE_ENTER(function_name) \
  rgdt_debug_report( \
    NULL, \
    RGDT_DEBUG_LEVEL_1, \
    function_name, \
    "Enter Function", \
    RGDT_RPT_INFO, \
    0, 0)

#define RGDT_TRACE_EXIT(function_name, ret_val) \
  rgdt_debug_report( \
    NULL, \
    RGDT_DEBUG_LEVEL_1, \
    function_name, \
    "Exit Function", \
    RGDT_RPT_INFO, \
    ret_val, 0)

/* Macros for Reading/Writing Registers */
/* Uses the inb/inw/inl and outb/outw/outl kernel services */

#define RGDT_READ_BYTE_REG(reg_offset) \
  inb(inst_info_addr->io_base_address + reg_offset)
#define RGDT_WRITE_BYTE_REG(reg_offset, reg_value) \
  outb(inst_info_addr->io_base_address + reg_offset, reg_value)

#define RGDT_READ_WORD_REG(reg_offset) \
  inw(inst_info_addr->io_base_address + reg_offset)
#define RGDT_WRITE_WORD_REG(reg_offset, reg_value) \
  outw(inst_info_addr->io_base_address + reg_offset, reg_value)

#define RGDT_READ_DWORD_REG(reg_offset) \
  inl(inst_info_addr->io_base_address + reg_offset)
#define RGDT_WRITE_DWORD_REG(reg_offset, reg_value) \
  outl(inst_info_addr->io_base_address + reg_offset, reg_value)

/******************************************************************************/
/*                              Type Definitions                              */
/******************************************************************************/

/* Ethernet Types */

typedef struct
{
  unsigned char address[RGDT_ETHER_ADDRESS_LEN];
} rgdt_ether_address_t, *rgdt_ether_address_addr_t;

/* Multicast List Types */

typedef struct rgdt_mclist_entry_t
{
  rgdt_ether_address_t mcast_address;
  struct rgdt_mclist_entry_t* next_addr;
} rgdt_mclist_entry_t, *rgdt_mclist_entry_addr_t;

typedef struct
{
  uint32_t count;
  rgdt_mclist_entry_addr_t head_addr;
} rgdt_mclist_t, rgdt_mclist_addr_t;

/*
 * TBD: Define the data structures/constants used for reading and writing the
 *      register on the device and for transferring data between the device
 *      and the driver.
 */


/* Structure used to manage Solaris DDI DMA objects */
typedef struct
{
  ddi_dma_handle_t dma_handle;    /* DMA handle for object */
  ddi_acc_handle_t access_handle; /* Data access handle for object */
  caddr_t virt_addr;              /* Kernel virtual address */
  uint32_t phys_addr;             /* Physical address (DMA-able) */
} rgdt_dma_object_t;


/* Device Statistics */

typedef struct
{
  uint64_t if_speed;                   /* Interface speed (bits-per-second) */
  uint32_t duplex;                     /* Duplex mode */
  uint32_t media;                      /* Current media selection */
  uint32_t intr;                       /* Interrupts received and claimed */
  uint32_t norcvbuf;                   /* Packets rcvd/dropped due to no buff */
  uint32_t ierrors;                    /* Total packets rcvd with errors */
  uint32_t oerrors;                    /* Total packets sent with errors */
  uint32_t missed;                     /* Packets rcvd/dropped by hardware */
  uint32_t uflo;                       /* Times FIFO underflowed on xmt */
  uint32_t oflo;                       /* Times receiver overflowed on rcv */
  uint32_t align_errors;               /* Packets rcvd with framing errors */
  uint32_t fcs_errors;                 /* Packets rcvd with CRC errors */
  uint32_t carrier_errors;             /* Carrier lost or not detected on xmt */
  uint32_t collisions;                 /* Ethernet collisions during xmt */
  uint32_t ex_collisions;              /* Frames with excessive collisions */
  uint32_t tx_late_collisions;         /* Late collisions ( > 512 bit times) */
  uint32_t defer_xmits;                /* First xmt delayed (medium busy) */
  uint32_t first_collisions;           /* Successful xmt with 1 collision */
  uint32_t multi_collisions;           /* Successful xmt with > 1 collisions */
  uint32_t sqe_errors;                 /* SQE test error reported */
  uint32_t macxmt_errors;              /* Packets with transmit MAC failure */
  uint32_t macrcv_errors;              /* Packets rcvd with MAC errors */
  uint32_t toolong_errors;             /* Packets rcvd larger than max */
  uint32_t runt_errors;                /* Packets rcvd smaller than min */
} rgdt_stats_t, *rgdt_stats_addr_t;

/* Per-instance Private Data */

typedef struct
{
  /* Device information */
  gld_mac_info_t* gld_info_ptr;        /* GLD MAC information */
  dev_info_t* dev_info_addr;           /* Solaris device information */
  uint8_t irq;                         /* Interrupt request level */
  uint32_t io_base_address;            /* Device I/O space address */
  uint32_t mem_base_address;           /* Device memory space address */
  rgdt_ether_address_t bcast_address;  /* Device broadcast address */
  rgdt_ether_address_t mac_address; /* Device physical address */
  rgdt_mclist_t mcast_list;         /* Multicast address list */
  rgdt_stats_t stats;               /* Device statistics */
  uint8_t rx_mode;                     /* Current receive mode */

  /* Driver information */
  ddi_iblock_cookie_t iblock_cookie;   /* Interrupt block cookie */

  /* Resource Locks */
  kmutex_t intr_mutex;                 /* Protect intr-side fields */

  /* Media Selection */
  uint8_t duplex_mode;                 /* Duplex requested */
  uint8_t duplex_lock;                 /* Lock duplex mode */
  uint8_t default_port;                /* Selected port */
  uint8_t media_lock;                  /* Do not sense media */
  uint8_t link_speed;                  /* Link speed (Mbps) */
  uint8_t speed_lock;                  /* Force link speed value */

  /*
   * TBD: Define the per-instance information specific to the device.
   *      This includes fields required for tracking transmit/receive status,
   *      DMA resources, link negotiation status, etc.
   */
} rgdt_inst_info_t, *rgdt_inst_info_addr_t;


/******************************************************************************/
/*                             Function Prototypes                            */
/******************************************************************************/

/* General Driver Service */

void rgdt_debug_report(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  int min_debug_level,                 /* Minimum debug level */
  char* caller_name_addr,              /* Address of caller name */
  char* debug_msg_addr,                /* Address of debug message */
  uint32_t msg_type,                   /* Message type identifier */
  int app_error_code,                  /* Application error code */
  int os_error_code);                  /* OS error code */

void rgdt_log_message(
  char* description_addr,              /* Address of description of message */
  mblk_t* message_addr);               /* Address of STREAMS message to log */

void rgdt_read_properties(
  dev_info_t* dev_info_addr);          /* Address of device info */

void rgdt_report_error(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  char* caller_name_addr,              /* Address of caller name */
  char* error_msg_addr,                /* Address of error message */
  uint32_t msg_type,                   /* Message type identifier */
  int app_error_code,                  /* Application error code */
  int os_error_code);                  /* OS error code */

/* Device Dependent Services */

int rgdt_attach_dd(
  rgdt_inst_info_addr_t inst_info_addr);/* Address of per-instance info */

int rgdt_detach_dd(
  rgdt_inst_info_addr_t inst_info_addr);/* Address of per-instance info */

void rgdt_disable_interrupts_dd(
  rgdt_inst_info_addr_t inst_info_addr);/* Address of per-instance info */

uint16_t rgdt_do_mii_autoneg_dd(
  rgdt_inst_info_addr_t inst_info_addr);/* Address of per-instance info */

void rgdt_enable_interrupts_dd(
  rgdt_inst_info_addr_t inst_info_addr);/* Address of per-instance info */

int rgdt_interrupt_dd(
  rgdt_inst_info_addr_t inst_info_addr);/* Address of per-instance info */

int16_t rgdt_mdio_read_dd(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  int32_t phy_id,                      /* PHY ID */
  int32_t location);                   /* Location in MII */

void rgdt_mdio_sync_dd(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  int bits);                           /* Number of bits to send */

void rgdt_mdio_write_dd(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  int32_t phy_id,                      /* PHY ID */
  int32_t location,                    /* Location in MII */
  int16_t value);                      /* Value to write to location */

int rgdt_probe_dd(
  dev_info_t* dev_info_addr);          /* Address of device info */

int rgdt_read_mac_address_dd(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  rgdt_ether_address_addr_t mac_address_addr); /* Address of buffer to */
                                               /* hold MAC address */

int rgdt_reset_dd(
  rgdt_inst_info_addr_t inst_info_addr);/* Address of per-instance info */

uint32_t rgdt_rx_interrupt_dd(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  uint16_t intr_status);               /* Status value read from device */

void rgdt_set_rx_mode_dd(
  rgdt_inst_info_addr_t inst_info_addr);/* Address of per-instance info */

int rgdt_start_dd(
  rgdt_inst_info_addr_t inst_info_addr);/* Address of per-instance info */

int rgdt_stop_dd(
  rgdt_inst_info_addr_t inst_info_addr);/* Address of per-instance info */

int rgdt_transmit_dd(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  mblk_t* message_ptr);                /* Pointer to STREAMS message block */

uint32_t rgdt_tx_interrupt_dd(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  uint16_t intr_status);               /* Status value read from device */

int rgdt_write_mac_address_dd(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  rgdt_ether_address_addr_t mac_address_addr); /* Address of buffer with */
                                               /* MAC address */

/* End Standard C Header File */

#if defined(__cplusplus)
}
#endif

#endif  /* !defined(_RGDT_H_) */
