/*******************************************************************************
*
* Name:
*
*   rgdt_dd.c  (Rice GLD Driver Template - Device Dependent)
*
* Author:
*
*   Michael Rice               2000
*   RiceM-Drivers@att.net
*
*   This software may be used and distributed according to the terms
*   of the GNU Public License, incorporated herein by reference.
*
* Description:
*   This file provides the device-dependent services required to implement the
*   network device driver for DEVDESC devices.
*
*
* Contents:
*
*   Device-dependent services:
*   rgdt_attach_dd                   - Device-dependent attach.      (Internal).
*   rgdt_detach_dd                   - Device-dependent detach.      (Internal).
*   rgdt_disable_interrupts_dd       - Disable interrupt delivery.   (Internal).
*   rgdt_do_mii_autoneg_dd           - Auto-negotiate a link.        (Internal).
*   rgdt_enable_interrupts_dd        - Enable interrupt delivery.    (Internal).
*   rgdt_interrupt_dd                - Process an interrupt.         (Internal).
*   rgdt_mdio_read_dd                - Read MII values.              (Internal).
*   rgdt_mdio_sync_dd                - Generate MII sync preamble.   (Internal).
*   rgdt_mdio_write_dd               - Write MII values.             (Internal).
*   rgdt_probe_dd                    - Device-dependent probe.       (Internal).
*   rgdt_read_mac_address_dd         - Read MAC address from device. (Internal).
*   rgdt_reset_dd                    - Device-dependent reset.       (Internal).
*   rgdt_rx_interrupt_dd             - Process a receive interrupt.  (Internal).
*   rgdt_set_rx_mode_dd              - Set the device receive mode.  (Internal).
*   rgdt_start_dd                    - Device-dependent start.       (Internal).
*   rgdt_stop_dd                     - Device-dependent stop.        (Internal).
*   rgdt_transmit_dd                 - Transmit data to the device.  (Internal).
*   rgdt_tx_interrupt_dd             - Process a transmit interrupt. (Internal).
*   rgdt_write_mac_address_dd        - Write MAC address to device.  (Internal).
*
* Additional Comments:
*
*   1. See the Solaris gld(7D) man page for general information on the 
*      Generic LAN Driver.
*   2. See the Solaris gld(9E) man page for information on the required entry 
*      points for GLD-based device drivers.
*   3. See the Solaris gld(9F) man page for information on the services 
*      provided by GLD for GLD-based device drivers.
*   4. See the Solaris gld_mac_info(9S) and gld_stats(9S) man pages for 
*      information about the GLD-specific data structures used by GLD-based
*      device drivers.
*   5. As of Solaris 8 (SunOS 5.8), the Generic LAN Driver is implemented only
*      for Intel Architecture (IA) systems.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
*******************************************************************************/

/*****************************************************************************/
/*                             Include Files                                 */
/*****************************************************************************/

#include <sys/types.h>                 /* System Type Definitions */
#include <sys/conf.h>                  /* Module Configuration Definitions */
#include <sys/debug.h>                 /* Debug Macro Definitions */
#include <sys/stropts.h>               /* STREAMS Options Definitions */
#include <sys/stream.h>                /* STREAMS Definitions */
#include <sys/strsun.h>                /* Sun STREAMS Definitions */
#include <sys/strlog.h>                /* STREAMS Log Definitions */
#include <sys/dlpi.h>                  /* Data Link Provider Interface Defns */
#include <sys/ethernet.h>              /* Ethernet Definitions */
#include <sys/gld.h>                   /* Generic LAN Driver Definitions */
#include <sys/kmem.h>                  /* Kernel Memory Definitions */
#include <sys/modctl.h>                /* Module Control Definitions */
#include <sys/errno.h>                 /* System Error Code Definitions */
#include <sys/ddi.h>                   /* Device Driver Interface Definitions */
#include <sys/sunddi.h>                /* Sun Device Driver Interface Defns */

#include "rgdt.h"                      /* Driver Definitions */

/*****************************************************************************/
/*                                Pragmas                                    */
/*****************************************************************************/

/*****************************************************************************/
/*                            Module Statics                                 */
/*****************************************************************************/

static struct ddi_device_acc_attr rgdt_dev_attr =
{
  DDI_DEVICE_ATTR_V0,
  DDI_STRUCTURE_LE_ACC,
  DDI_STRICTORDER_ACC
};

/******************************************************************************
*
* Name:
*
*   rgdt_attach_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service performs the device-dependent part of the attach().  It is 
*   called by the attach() service.  It does all of the setup for the device
*   and for the driver to get it ready to start processing interrupts.
*
*
* Inputs:
*
*   inst_info_addr - Address of driver per-instance information.
*
* Outputs:
*
*   inst_info_addr - Address of driver per-instance information.
*
* Return Values:
*
*   DDI_SUCCESS : The attach was completed successfully.
*   DDI_FAILURE : The attach failed.
*
* Additional Comments:
*
*   1. The probe() has already been completed, so we are certain that we are
*      dealing with the correct device type.
*   2. This service does not enable the device to generate interrupts, that is
*      done in the start_dd() service.
*   3. The DMA resources, Rx and Tx ring descriptors and buffers, are allocated
*      here.  The Rx and Tx ring descriptor commands are initialized in the
*      start_dd() service.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

int rgdt_attach_dd(
  rgdt_inst_info_addr_t inst_info_addr) /* Address of per-instance info */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;
	ddi_acc_handle_t conf_handle;
  char debug_msg[1024];

/*****************************************************************************/
/*                            Begin rgdt_attach_dd                           */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_ATTACH_DD);

  /*
   * Do stuff like get the MAC address from the board (required), 
   * setup the hardware, etc.
   *
   * Note: Don't enable interrupts, this is handled in the start service.
   */

  /* Map in the PCI configuration space */
  if(pci_config_setup(
    inst_info_addr->dev_info_addr,
    &conf_handle) == DDI_SUCCESS)
  {
    /* Get the I/O base address */
    inst_info_addr->io_base_address = pci_config_get32(
      conf_handle,
      RGDT_PCICFG_IO_BASE_ADDRESS);
    inst_info_addr->io_base_address &= ~3;

    /* Get the memory base address */
    inst_info_addr->mem_base_address = pci_config_get32(
      conf_handle,
      RGDT_PCICFG_MEM_BASE_ADDRESS);

    /* Get the device IRQ */
    inst_info_addr->irq = pci_config_get8(
      conf_handle,
      RGDT_PCICFG_INT_LINE);

    /* Unmap the PCI configuration space */
    pci_config_teardown(&conf_handle);

    /* Read the OEM node address */
    rgdt_read_mac_address_dd(
      inst_info_addr,
      &(inst_info_addr->mac_address));

    sprintf(debug_msg,
     "MAC Address: %02x:%02x:%02x:%02x:%02x:%02x, I/O Address: 0x%04x, IRQ: %d",
      inst_info_addr->mac_address.address[0],
      inst_info_addr->mac_address.address[1],
      inst_info_addr->mac_address.address[2],
      inst_info_addr->mac_address.address[3],
      inst_info_addr->mac_address.address[4],
      inst_info_addr->mac_address.address[5],
      inst_info_addr->io_base_address,
      inst_info_addr->irq);
    rgdt_debug_report(
      inst_info_addr,
      RGDT_DEBUG_LEVEL_2, RGDT_ATTACH_DD,
      debug_msg,
      RGDT_RPT_INFO, 0, 0);

    /* Reset the device */
    rgdt_reset_dd(inst_info_addr);

    /* TBD: Initialize the media selection params (link speed, duplex, etc.) */

    /*
     * TBD: Allocate the transmit/receive buffers and rings.  Initialize the
     *      ring pointers, transmit/receive status indicators, etc. in the
     *      per-instance structure.
     */

    completion_status = DDI_SUCCESS;
  }
  else  /* Error mapping configuration space */
  {
    /* Report the error */
    rgdt_report_error(
      inst_info_addr,
      RGDT_ATTACH_DD,
      "Map PCI config space failed",
      RGDT_RPT_ERROR,
      0, 0);

    completion_status = DDI_FAILURE;
  }

  RGDT_TRACE_EXIT(RGDT_ATTACH_DD, completion_status);

  return completion_status;
}  /* End rgdt_attach_dd */

/******************************************************************************
*
* Name:
*
*   rgdt_detach_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service performs the device-dependent part of the detach().  It is 
*   called by the detach() service.  It deallocates all of the DMA resources
*   that were allocated for the device.
*
*
* Inputs:
*
*   inst_info_addr - Address of driver per-instance information.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   DDI_SUCCESS : The detach was completed successfully.
*   DDI_FAILURE : The detach failed.
*
* Additional Comments:
*
*   1. All of the DMA resources, Rx and Tx ring descriptors and buffers, are
*      deallocated here.
*   2. It is assumed that the Rx and Tx channels on the device have already
*      been stopped, such that the device is no longer performing any DMA
*      operations.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

int rgdt_detach_dd(
  rgdt_inst_info_addr_t inst_info_addr)/* Address of per-instance info */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;

/*****************************************************************************/
/*                            Begin rgdt_detach_dd                           */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_DETACH_DD);

  /* Reset the device */
  rgdt_reset_dd(inst_info_addr);

  /*
   * TBD: Deallocate all of the buffers/rings that were allocated in
   *      the rgdt_attach_dd service.
   */

  completion_status = DDI_SUCCESS;

  RGDT_TRACE_EXIT(RGDT_DETACH_DD, completion_status);

  return completion_status;
}  /* End rgdt_detach_dd */

/******************************************************************************
*
* Name:
*
*   rgdt_disable_interrupts_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service disables interrupt delivery from the device to the host.
*
*
* Inputs:
*
*   inst_info_addr - Address of driver per-instance information.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   None.
*
* Additional Comments:
*
*   None.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

void rgdt_disable_interrupts_dd(
  rgdt_inst_info_addr_t inst_info_addr)/* Address of per-instance info */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

/*****************************************************************************/
/*                      Begin rgdt_disable_interrupts_dd                     */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_DISABLE_INTERRUPTS_DD);

  /* TBD: Disable interrupt delivery */

  RGDT_TRACE_EXIT(RGDT_DISABLE_INTERRUPTS_DD, 0);
}  /* End rgdt_disable_interrupts_dd */

/******************************************************************************
*
* Name:
*
*   rgdt_do_mii_autoneg_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service auto-negotiates a link for the PHY using the MII.
*
*
* Inputs:
*
*   inst_info_addr - Address of driver per-instance information.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   RGDT_SUCCESS : Auto-negotiation completed and the link status is valid.
*   RGDT_FAILURE : Either auto-negotiation did not complete or 
*                     there is no valid link status.
*
* Additional Comments:
*
*   1. This service assumes that a MII PHY was detected on the device during the
*      device-dependent attach() service.
*   2. This service only applies to devices that use a 
*      Media Independent Interface (MII) for managing the physical link.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

uint16_t rgdt_do_mii_autoneg_dd(
  rgdt_inst_info_addr_t inst_info_addr)/* Address of per-instance info */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  uint16_t completion_status;
  char debug_msg[1024];

/*****************************************************************************/
/*                        Begin rgdt_do_mii_autoneg_dd                       */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_DO_MII_AUTONEG_DD);

  completion_status = RGDT_FAILURE;

  /* Disable device interrupts */
  rgdt_disable_interrupts_dd(inst_info_addr);

  /* TBD: Do what needs to be done to tell the MII to negotiate a link */

  /* Re-enable device interrupts */
  rgdt_enable_interrupts_dd(inst_info_addr);

  RGDT_TRACE_EXIT(RGDT_DO_MII_AUTONEG_DD, completion_status);

  return completion_status;
}  /* End rgdt_do_mii_autoneg_dd */

/******************************************************************************
*
* Name:
*
*   rgdt_enable_interrupts_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service enables interrupt delivery from the device to the host.
*
*
* Inputs:
*
*   inst_info_addr - Address of driver per-instance information.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   None.
*
* Additional Comments:
*
*   None.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

void rgdt_enable_interrupts_dd(
  rgdt_inst_info_addr_t inst_info_addr)/* Address of per-instance info */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

/*****************************************************************************/
/*                      Begin rgdt_enable_interrupts_dd                      */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_ENABLE_INTERRUPTS_DD);

  /* TBD: Enable interrupt delivery */

  RGDT_TRACE_EXIT(RGDT_ENABLE_INTERRUPTS_DD, 0);
}  /* End rgdt_enable_interrupts_dd */

/******************************************************************************
*
* Name:
*
*   rgdt_interrupt_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service handles a potential interrupt by the device.  It checks if the
*   device actually interrupted and, if necessary, processes the interrupt.
*
*
* Inputs:
*
*   inst_info_addr - Address of driver per-instance information.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   DDI_INTR_CLAIMED   : The device interrupted and the interrupt was processed.
*   DDI_INTR_UNCLAIMED : The device did not interrupt.
*
* Additional Comments:
*
*   1. This service continues processing interrupts until either the device
*      stops interrupting or the maximum number of events has been reached.
*      The "maximum ISR events" check prevents this ISR from monopolizing the
*      system during a heavy loading period.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

int rgdt_interrupt_dd(
  rgdt_inst_info_addr_t inst_info_addr)/* Address of per-instance info */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;
  uint16_t intr_status;
  uint16_t intr_enabled;
  uint32_t events_processed;

/*****************************************************************************/
/*                          Begin rgdt_interrupt_dd                          */
/*****************************************************************************/

  /*
   * TBD: A generic interrupt processing architecture is shown here.  This will,
   *      in all probability, need to be modified for each specific device.
   */

  RGDT_TRACE_ENTER(RGDT_INTERRUPT_DD);

  events_processed = 0;

  /*
   * TBD: Check if interrupts are enabled and if this device interrupted.
   */
  intr_enabled = 0;
  intr_status = 0;
  if((intr_enabled) && (intr_status != 0))
  {
    /*
     * Disable interrupt delivery from this device to the host.  Interrupts
     * will be re-enabled by rgdt_gldm_intr().
     */
    rgdt_disable_interrupts_dd(inst_info_addr);

    /*
     * Increment the events processed counter so we know we processed at least
     * one interrupt.  This is done so that when we exit the interrupt 
     * processing loop, we will correctly acknowledge that this was an interrupt
     * from our device.  If we didn't increment the event counter here, and no
     * work is done in either the receive or transmit interrupt handlers, then
     * this service would not acknowledge the interrupt and rgdt_gldm_intr()
     * would not re-enable interrupt delivery from the device to the host.
     */
    events_processed++;
  }

  /*
   * Process interrupt events until there are no more or until the maximum
   * number of events has been reached.
   *
   * The ISR will only process events up to a driver-specified maximum count.
   * This is done to give some CPU time to other drivers in the system when
   * this device is heavily loaded.
   */
  while((intr_status != 0) && (events_processed < RGDT_MAX_ISR_EVENTS))
  {
    /* TBD: Get the interrupt status from the device and clear it */
    intr_status = 0;

    /* TBD: Check if an interrupt actually occurred */
    if(intr_status != 0)
    {
      rgdt_debug_report(
        inst_info_addr,
        RGDT_DEBUG_LEVEL_2, RGDT_INTERRUPT_DD,
        "Processing interrupt, Status =>",
        RGDT_RPT_INFO, intr_status, 0);

      /* TBD: Check if there are packets to receive */
      if(intr_status)
      {
        /* Process the received packet(s) */
        events_processed += rgdt_rx_interrupt_dd(
          inst_info_addr, intr_status);

        /* TBD: May need to wakeup the DMA engine on the device */
      }

      /* TBD: Check if a transmit has completed or there are transmit errors */
      if(intr_status)
      {
        /* Process the transmitted packets */
        events_processed += rgdt_tx_interrupt_dd(
          inst_info_addr, intr_status);
      }

      /* TBD: Check for error interrupts */
      if(intr_status)
      {
        /* Report the error */
        rgdt_report_error(
          inst_info_addr,
          NULL,
          "Error interrupt, Status =>",
          RGDT_RPT_ERROR,
          intr_status, 0);

        /*
         * TBD: Check what type of error happened (collisions, CRCs, etc.) and
         *      updated the stats per-instance structure.
         *
         *      Also, if this is a fatal error (PCI fault, etc.) then may need
         *      to reset the device and the transmit/receive buffers and rings.
         */

        /* Increment the event counter */
        events_processed++;
      }

      /* Increment the interrupt count */
      inst_info_addr->stats.intr++;

    }  /* End if interrupt status != 0 */
  }  /* End while device interrupting and max ISR events not exceeded */

  /* Check if any events were processed */
  if(events_processed > 0)
  {
    /* Check if we were unable to process all events */
    if(events_processed >= RGDT_MAX_ISR_EVENTS)
    {
      rgdt_report_error(
        inst_info_addr,
        NULL,
        "Did not process all events, last interrupt status =>",
        RGDT_RPT_WARNING,
        intr_status, 0);
    }

    /*
     * Set the completion status to indicate this was indeed an interrupt from
     * our device, which will prevent the system from passing the interrupt to
     * any other device which may be sharing the IRQ.
     */
    completion_status = DDI_INTR_CLAIMED;
  }
  else  /* Our device did not interrupt */
  {
    /*
     * Set the completion status to indicate this was not an interrupt from
     * our device, which allows the system to pass the interrupt to another
     * device which may be sharing the IRQ.
     */
    completion_status = DDI_INTR_UNCLAIMED;
  }

  RGDT_TRACE_EXIT(RGDT_INTERRUPT_DD, completion_status);

  return completion_status;
}  /* End rgdt_interrupt_dd */

/******************************************************************************
*
* Name:
*
*   rgdt_mdio_read_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service uses the MII management interface to read data from the MII.
*
*
* Inputs:
*
*   inst_info_addr - Address of driver per-instance information.
*   phy_id         - PHY identifier.
*   location       - Location in MII to read data from.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   int16_t : Value read from MII.
*
* Additional Comments:
*
*   1. This service assumes the MII Management Data I/O (MDIO) controller is
*      idle and does not return until the command is finished.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

int16_t rgdt_mdio_read_dd(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  int32_t phy_id,                      /* PHY ID */
  int32_t location)                    /* Location in MII */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  uint16_t mii_value;

/*****************************************************************************/
/*                          Begin rgdt_mdio_read_dd                          */
/*****************************************************************************/

  mii_value = 0;

  /* TBD: Read the specified location using the MII */
  
	return mii_value;
}  /* End rgdt_mdio_read_dd */

/******************************************************************************
*
* Name:
*
*   rgdt_mdio_sync_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service uses the MII management interface to generate the preamble
*   required for initial synchronization.
*
*
* Inputs:
*
*   inst_info_addr - Address of driver per-instance information.
*   bits           - Number of bits to send to try and get synch.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   None.
*
* Additional Comments:
*
*   None.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

void rgdt_mdio_sync_dd(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  int bits)                            /* Number of bits to send */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

/*****************************************************************************/
/*                          Begin rgdt_mdio_sync_dd                          */
/*****************************************************************************/

  /* TBD: Write the specified number bits to synchronize the MII */
  
}  /* End rgdt_mdio_sync_dd */

/******************************************************************************
*
* Name:
*
*   rgdt_mdio_write_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service uses the MII management interface to write data to the MII.
*
*
* Inputs:
*
*   inst_info_addr - Address of driver per-instance information.
*   phy_id         - PHY identifier.
*   location       - Location in MII to read data from.
*   value          - Value to write to specified location.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   None.
*
* Additional Comments:
*
*   1. This service assumes the MII Management Data I/O (MDIO) controller is
*      idle and does not return until the command is finished.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

void rgdt_mdio_write_dd(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  int32_t phy_id,                      /* PHY ID */
  int32_t location,                    /* Location in MII */
  int16_t value)                       /* Value to write to location */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

/*****************************************************************************/
/*                          Begin rgdt_mdio_write_dd                         */
/*****************************************************************************/

  /* TBD: Write the specified value to the specified location using the MII */
  
}  /* End rgdt_mdio_write_dd */

/******************************************************************************
*
* Name:
*
*   rgdt_probe_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service does the device-dependent part of the probe().  It accesses the
*   device to determine if the device is actually present in the system.
*
*
* Inputs:
*
*   dev_info_addr - Address of per-instance device information maintained by OS.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   DDI_PROBE_SUCCESS  : Probe was successful.
*   DDI_PROBE_FAILURE  : The probe failed.
*   DDI_PROBE_DONTCARE : Probe was unsuccessful, but attach() should be called.
*   DDI_PROBE_PARTIAL  : Device not present now, but may be in the future.
*
* Additional Comments:
*
*   None.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

int rgdt_probe_dd(
  dev_info_t* dev_info_addr)           /* Address of device info */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;
  int reg_index;
	ddi_acc_handle_t conf_handle;
  uint32_t conf_data;
  uint16_t vendor_id;
  uint16_t device_id;

/*****************************************************************************/
/*                            Begin rgdt_probe_dd                            */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_PROBE_DD);

  /* Map in the PCI configuration space */
  if(pci_config_setup(dev_info_addr, &conf_handle) == DDI_SUCCESS)
  {
    /* Dump the PCI configuration space to the system log */
    rgdt_debug_report(
      NULL,
      RGDT_DEBUG_LEVEL_3,
      RGDT_PROBE_DD,
      "Dumping config address space (256 bytes, 64 dwords):",
      RGDT_RPT_INFO,
      0, 0);
    for(reg_index = 0; reg_index < 64; reg_index++)
    {
      conf_data = pci_config_get32(conf_handle, reg_index * 4);
      rgdt_debug_report(
        NULL,
        RGDT_DEBUG_LEVEL_3,
        RGDT_PROBE_DD,
        "Config data word",
        RGDT_RPT_INFO,
        reg_index, conf_data);
    }

    /* Read the vendor ID and device ID */
    vendor_id = pci_config_get16(conf_handle, RGDT_PCICFG_VENDOR_ID);
    device_id = pci_config_get16(conf_handle, RGDT_PCICFG_DEVICE_ID);

    /* Check to see if this is the right device type */
    if((vendor_id == RGDT_VENDOR_ID) && (device_id == RGDT_DEVICE_ID))
    {
      rgdt_debug_report(
        NULL,
        RGDT_DEBUG_LEVEL_1, RGDT_PROBE_DD,
        "Probe succeeded: Vendor ID and Device ID",
        RGDT_RPT_INFO, vendor_id, device_id);

      completion_status = DDI_PROBE_SUCCESS;
    }
    else  /* Not the expected device type */
    {
      rgdt_debug_report(
        NULL,
        RGDT_DEBUG_LEVEL_1, RGDT_PROBE_DD,
        "Probe failed: Vendor ID and Device ID",
        RGDT_RPT_WARNING, vendor_id, device_id);

      completion_status = DDI_PROBE_FAILURE;
    }

    /* Unmap the PCI configuration space */
    pci_config_teardown(&conf_handle);
  }
  else  /* Error mapping configuration space */
  {
    /* Report the error */
    rgdt_report_error(
      NULL,
      RGDT_PROBE_DD,
      "Map PCI config space failed",
      RGDT_RPT_ERROR,
      0, 0);

    completion_status = DDI_PROBE_FAILURE;
  }

  RGDT_TRACE_EXIT(RGDT_PROBE_DD, completion_status);

  return completion_status;
}  /* End rgdt_probe_dd */

/******************************************************************************
*
* Name:
*
*   rgdt_read_mac_address_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service reads the MAC address from the device.  The address read is the
*   OEM node address of the device, not the address that is programmed into the
*   device by the rgdt_write_mac_address_dd() service.  If the mac-address
*   property is defined in the driver configuration file, the MAC address is
*   not read from the device.
*
*
* Inputs:
*
*   inst_info_addr   - Address of driver per-instance information.
*   mac_address_addr - Address of buffer to receive MAC address read either
*                      from the driver configuration file or from the device.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   DDI_SUCCESS  : Address read successfully.
*   DDI_FAILURE  : Error reading address.
*
* Additional Comments:
*
*   None.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

int rgdt_read_mac_address_dd(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  rgdt_ether_address_addr_t mac_address_addr)  /* Address of buffer to */
                                               /* hold MAC address */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;
  int read_device;
  int* int_prop_values_ptr;
  int num_prop_values;

/*****************************************************************************/
/*                       Begin rgdt_read_mac_address_dd                      */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_READ_MAC_ADDRESS_DD);

  /* Initialize the flag to not read the MAC address from the device */
  read_device = 0;

  /* Read the mac-address driver property */
  if(ddi_prop_lookup_int_array(
    DDI_DEV_T_ANY,
    inst_info_addr->dev_info_addr,
    DDI_PROP_DONTPASS,
    "mac-address",
    &int_prop_values_ptr,
    &num_prop_values) == DDI_PROP_SUCCESS)
  {
    /* Make sure three values were specified */
    if(num_prop_values == 3)
    {
      /* Save the specified MAC address */
      ((uint16_t*)(mac_address_addr->address))[0] = 
        int_prop_values_ptr[0] & 0xFFFF;
      ((uint16_t*)(mac_address_addr->address))[1] = 
        int_prop_values_ptr[1] & 0xFFFF;
      ((uint16_t*)(mac_address_addr->address))[2] = 
        int_prop_values_ptr[2] & 0xFFFF;
    }
    else  /* Invalid mac-address property */
    {
      /* Report the error */
      rgdt_report_error(
        inst_info_addr,
        NULL,
        "Invalid mac-address property definition, number of values read =>",
        RGDT_RPT_WARNING,
        num_prop_values, 0);

      /* Set the flag to read the MAC address from the device */
      read_device = 1;
    }

    /* Deallocate the property memory */
    ddi_prop_free(int_prop_values_ptr);
  }
  else  /* Property not defined */
  {
    /* Set the flag to read the MAC address from the device */
    read_device = 1;
  }
  
  /* Check if the MAC address should be read from the device */
  if(read_device)
  {
    /*
     * TBD: Read the MAC address from the device and put the value into the
     *      mac_address_addr buffer.
     */
    ((uint16_t*)(mac_address_addr->address))[0] = 0x1234;
    ((uint16_t*)(mac_address_addr->address))[1] = 0x5678;
    ((uint16_t*)(mac_address_addr->address))[2] = 0x9ABC;
  }
  
  completion_status = DDI_SUCCESS;

  RGDT_TRACE_EXIT(RGDT_READ_MAC_ADDRESS_DD, completion_status);

  return completion_status;
}  /* End rgdt_read_mac_address_dd */

/******************************************************************************
*
* Name:
*
*   rgdt_reset_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service resets the device to a known good initial state.
*
*
* Inputs:
*
*   inst_info_addr - Address of driver per-instance information.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   DDI_SUCCESS  : Device reset successfully.
*   DDI_FAILURE  : Error resetting the device.
*
* Additional Comments:
*
*   1. This service is also called from the attach_dd() service, prior to 
*      configuring the internal Tx/Rx rings or the DMA buffers in the 
*      per-instance stucture.  Do not try to access those buffers in this 
*      service.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

int rgdt_reset_dd(
  rgdt_inst_info_addr_t inst_info_addr)/* Address of per-instance info */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;

/*****************************************************************************/
/*                            Begin rgdt_reset_dd                            */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_RESET_DD);

  /* TBD: Reset the device to a known good state. */

  completion_status = DDI_SUCCESS;

  RGDT_TRACE_EXIT(RGDT_RESET_DD, completion_status);

  return completion_status;
}  /* End rgdt_reset_dd */

/******************************************************************************
*
* Name:
*
*   rgdt_rx_interrupt_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service processes receive-related interrupts, such as
*   receive-complete, receive-early, etc., from the device.
*
*
* Inputs:
*
*   inst_info_addr - Address of driver per-instance information.
*   intr_status    - Interrupt status value read from the device.
*
* Outputs:
*
*   Valid packets are copied to STREAMS messages containing the packet data
*   and sent up to GLD.
*
* Return Values:
*
*   uint32_t : Number of packets processed.
*
* Additional Comments:
*
*   None.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

uint32_t rgdt_rx_interrupt_dd(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  uint16_t intr_status)                /* Status value read from device */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  uint32_t completion_status;

/*****************************************************************************/
/*                         Begin rgdt_rx_interrupt_dd                        */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_RX_INTERRUPT_DD);

  completion_status = 0;

  /*
   * TBD: Process all of the packets that have been received by the device
   *      (and copied to the host buffers if using DMA).
   *
   *      For each received packet, copy it into a new STREAMS message block
   *      and send the STREAMS message block to GLD using the gld_recv service.
   *
   *      Example:
   *        uint32_t packet_len;
   *        mblk_t* message_ptr;
   *
   *        FOREACH Received Packet
   *          pkt_ptr = <Get the buffer address from the receive ring>
   *          packet_len = <Get the packet length for the received packet>
   *          message_ptr = allocb(packet_len, BPRI_LO);
   *          if(message_ptr != NULL)
   *          {
   *            bcopy((caddr_t)pkt_ptr, (caddr_t)message_ptr->b_rptr, packet_len);
   *            message_ptr->b_wptr += packet_len;
   *            gld_recv(inst_info_addr->gld_info_ptr, message_ptr);
   *            completion_status++;
   *          }
   *          else
   *          {
   *            <Report the very serious memory allocation failure>
   *          }
   *        END FOREACH
   */

  RGDT_TRACE_EXIT(RGDT_RX_INTERRUPT_DD, completion_status);

  return completion_status;
}  /* End rgdt_rx_interrupt_dd */

/******************************************************************************
*
* Name:
*
*   rgdt_set_rx_mode_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service sets the receive mode (promiscuous, multicast, etc.) for the
*   device.
*
*
* Inputs:
*
*   inst_info_addr - Address of driver per-instance information.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   None.
*
* Additional Comments:
*
*   None.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

void rgdt_set_rx_mode_dd(
  rgdt_inst_info_addr_t inst_info_addr)/* Address of per-instance info */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

/*****************************************************************************/
/*                         Begin rgdt_set_rx_mode_dd                         */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_SET_RX_MODE_DD);

  /*
   * TBD: Update the registers in the device to indicate what packets should be
   *      accepted by the MAC.  May also need to update the device with a new
   *      multicast address table.  The new receive mode and the current 
   *      multicast address list are in the per-instance structure.
   */

  RGDT_TRACE_EXIT(RGDT_SET_RX_MODE_DD, 0);
}  /* End rgdt_set_rx_mode_dd */

/******************************************************************************
*
* Name:
*
*   rgdt_start_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service tries to auto-negotiate the link parameters with the link
*   partner.  This service also enables the device to generate interrupts and 
*   prepares the driver to transmit/receive data to/from the device.
*
*
* Inputs:
*
*   inst_info_addr - Address of driver per-instance information.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   DDI_SUCCESS : The operation completed successfully.
*   DDI_FAILURE : A hardware error occurred.
*
* Additional Comments:
*
*   None.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

int rgdt_start_dd(
  rgdt_inst_info_addr_t inst_info_addr)/* Address of per-instance info */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;

/*****************************************************************************/
/*                            Begin rgdt_start_dd                            */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_START_DD);

  completion_status = DDI_SUCCESS;

  /* TBD: Auto-negotiate the link */

  /*
   * TBD: Initialize the transmit/receive ring control information in the
   *      per-instance structure.  The buffers and rings were allocated in the
   *      rgdt_attach_dd service, this is just getting the ready to be used.
   */

  /* TBD: Prepare the device to start processing interrupts and data. */

  /* TBD: Enable interrupts on the device */

  RGDT_TRACE_EXIT(RGDT_START_DD, completion_status);

  return completion_status;
}  /* End rgdt_start_dd */

/******************************************************************************
*
* Name:
*
*   rgdt_stop_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service disables the device from generating interrupts and stops
*   the driver from processing received data.
*
*
* Inputs:
*
*   inst_info_addr - Address of driver per-instance information.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   DDI_SUCCESS : The operation completed successfully.
*   DDI_FAILURE : A hardware error occurred.
*
* Additional Comments:
*
*   None.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

int rgdt_stop_dd(
  rgdt_inst_info_addr_t inst_info_addr)/* Address of per-instance info */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;

/*****************************************************************************/
/*                            Begin rgdt_stop_dd                             */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_STOP_DD);

  /* Stop interrupt delivery to the host */
  rgdt_disable_interrupts_dd(inst_info_addr);

  /* TBD: Do whatever is necessary to stop the device from processing data */

  completion_status = DDI_SUCCESS;

  RGDT_TRACE_EXIT(RGDT_STOP_DD, completion_status);

  return completion_status;
}  /* End rgdt_stop_dd */

/******************************************************************************
*
* Name:
*
*   rgdt_transmit_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service transmits the specified STREAMS message block to the device.
*   If there are available buffers, the data is copied to the next available
*   transmit DMA buffer.  If no buffer space is available, no action is
*   performed and the data in the STREAMS message block is not copied.
*
*
* Inputs:
*
*   inst_info_addr - Address of driver per-instance information.
*   message_ptr    - Pointer to STREAMS message block containing data to
*                    transmit to the device.
*
* Outputs:
*
*   If the specified STREAMS message contains a valid packet, the packet data
*   is copied to a Tx ring buffer and queued to the device.
*
* Return Values:
*
*   DDI_SUCCESS  : Data transmitted successfully.
*   DDI_FAILURE  : No buffer space available, try again later.
*
* Additional Comments:
*
*   1. If there is no space in the Tx ring, the specified packet is not copied
*      and GLD is informed that there are no buffers available.  GLD will queue
*      the packet for a later retry.  When a transmit complete interrupt is 
*      processed, the Tx ISR informs GLD that it may retry any packets that it 
*      had previously queued.
*   2. The specified STREAMS message may be shorter than the minimum packet
*      length.  Either the device must automatically pad to the required minimum
*      length or it must be manually padded here, prior to sending it.
*   3. The specified STREAMS message does not contain the CRC value that must
*      be appended to all outgoing Ethernet frames.  Either the device must be
*      configured to automatically calculate and append the CRC to the frame
*      or it must be calculated and appended here, prior to sending it.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

int rgdt_transmit_dd(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  mblk_t* message_ptr)                 /* Pointer to STREAMS message block */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;

/*****************************************************************************/
/*                           Begin rgdt_transmit_dd                          */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_TRANSMIT_DD);

  completion_status = DDI_SUCCESS;

  /*
   * Disable interrupt delivery from the device to the host. This is done
   * to prevent concurrent access by this service and the Tx ISR to the Tx ring
   * buffers and indices.
   */
  rgdt_disable_interrupts_dd(inst_info_addr);

  /* TBD: Pad the frame if necessary */

  /* TBD: Calculate and append CRC for the frame, if ncessary */

  /* TBD: Send the frame to the device for transmission */

  /* Re-enable interrupt delivery from the device to the host */
  rgdt_enable_interrupts_dd(inst_info_addr);

  RGDT_TRACE_EXIT(RGDT_TRANSMIT_DD, completion_status);

  return completion_status;
}  /* End rgdt_transmit_dd */

/******************************************************************************
*
* Name:
*
*   rgdt_tx_interrupt_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service processes transmit-related interrupts, such as
*   tranmsit-complete, from the device.  It checks the transmit status 
*   and updates the device statistics maintained by the driver.
*
*
* Inputs:
*
*   inst_info_addr - Address of driver per-instance information.
*   intr_status    - Interrupt status value read from the device.
*
* Outputs:
*
*   None.
*
* Return Values:
*
*   uint32_t : Number of packets processed.
*
* Additional Comments:
*
*   None.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

uint32_t rgdt_tx_interrupt_dd(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  uint16_t intr_status)                /* Status value read from device */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  uint32_t completion_status;

/*****************************************************************************/
/*                         Begin rgdt_tx_interrupt_dd                        */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_TX_INTERRUPT_DD);

  completion_status = 0;

  /*
   * TBD: Process each pending transmit-related interrupt, incrementing the
   *      completion_completion status for each "chunk" of work performed.
   */

  /* Check if any work was completed */
  if(completion_status > 0)
  {
    /*
     * Notify GLD that it can retry any transmits that were previously queued
     * due to lack of space in the Tx ring.
     *
     * If GLD does not have any packets queued for transmit, it will ignore this
     * call.
     */
    gld_sched(inst_info_addr->gld_info_ptr);
  }

  RGDT_TRACE_EXIT(RGDT_TX_INTERRUPT_DD, completion_status);

  return completion_status;
}  /* End rgdt_tx_interrupt_dd */

/******************************************************************************
*
* Name:
*
*   rgdt_write_mac_address_dd
*
* Author:
*
*   Michael Rice               2000
*
* Description:
*
*   This service writes the MAC address to the device.
*
*
* Inputs:
*
*   inst_info_addr   - Address of driver per-instance information.
*   mac_address_addr - Address of MAC address to write to the device.
*
* Outputs:
*
*   The specified MAC address is written to the Station Address Register on the
*   device.
*
* Return Values:
*
*   DDI_SUCCESS : The specified MAC address was written into the device.
*   DDI_FAILURE : Error writing address.
*
* Additional Comments:
*
*   None.
*
* Revision History:
*   10 July 2000 - Michael Rice, RiceM-Drivers@att.net
*     - Original
*
******************************************************************************/

/*****************************************************************************/
/*                                 Function                                  */
/*****************************************************************************/

int rgdt_write_mac_address_dd(
  rgdt_inst_info_addr_t inst_info_addr,/* Address of per-instance info */
  rgdt_ether_address_addr_t mac_address_addr)  /* Address of buffer with */
                                               /* MAC address */
{

/*****************************************************************************/
/*                              Local Variables                              */
/*****************************************************************************/

  int completion_status;

/*****************************************************************************/
/*                      Begin rgdt_write_mac_address_dd                      */
/*****************************************************************************/

  RGDT_TRACE_ENTER(RGDT_WRITE_MAC_ADDRESS_DD);

  /* TBD: Write the specified Ethernet address into the device registers. */

  completion_status = DDI_SUCCESS;

  RGDT_TRACE_EXIT(RGDT_WRITE_MAC_ADDRESS_DD, completion_status);

  return completion_status;
}  /* End rgdt_write_mac_address_dd */
